#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2012-2013,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'

from __future__ import absolute_import, print_function
import csv
import logging
import os
import pwd
import smtplib
import socket
import time

from collections import defaultdict
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import gc3libs
import gc3libs.poller as plr
from gc3libs.cmdline import SessionBasedDaemon
from gc3libs.workflow import SequentialTaskCollection
from gc3libs import Run
from gc3libs.url import Url
from gc3libs.quantity import Duration

log = logging.getLogger('gc3.gc3utils')


class BemoviCSVConf(object):
    def __init__(self, path, default_walltime='3h'):
        self.path = path
        self.cfg = {}
        self.defaults = {}
        with open(path, 'rU') as fd:
            log.debug("Reading CSV configuration file %s", path)
            cr = csv.reader(fd)
            lineno = 0
            for line in cr:
                lineno += 1
                if len(line) != 9:
                    log.warning("Ignoring line '%d' in csv configuration file %s: wrong number of fields (%d != 9)",
                    lineno, path, len(line))
                    continue
                if line[0] in self.cfg:
                    log.warning(
                        "Overwriting dupliacate key in '%s' csv configuration file: '%s'",
                        csvcfgfile, line[0])

                try:
                    # Check if this is an header line. These values
                    # should always be integers.
                    int(line[1])
                    int(line[3])
                    int(line[4])
                    int(line[5])
                except ValueError:
                    log.debug("Ignoring line '%d' of file %s, some values do not convert to integer as expected.",
                              lineno, path)
                    continue
                data = {
                    'fps': line[1],
                    'pixel_to_scale': line[2],
                    'difference_lag': line[3],
                    'threshold1': line[4],
                    'threshold2': line[5],
                    'video_is_needed': False if line[7].lower() == 'optional' else True,
                    'email_to': line[8],
                }
                try:
                    data['requested_walltime'] = Duration(line[6])
                except ValueError as ex:
                    log.error("Unable to parse walltime '%s' for key %s in file"
                              " %s: %s. Using default value of %s",
                              line[6], line[0], path, ex, default_walltime)
                    data['requested_walltime'] = Duration(default_walltime)
                key = line[0]
                if key.lower() == "default":
                    self.defaults = data
                else:
                    self.cfg[key] = data

    def get(self, key):
        """Return configuration key for `path` entry, or default value (if any"""
        if key in self.cfg:
            return self.cfg[key].copy()
        elif os.path.basename(key) in self.cfg:
            return self.cfg[os.path.basename(key)].copy()
        else:
            return self.defaults.copy()

    def keys(self):
        return ['default'] + self.cfg.keys()

    def __iter__(self):
        return iter(self.keys())

class BemoviApp(gc3libs.Application):
    def update_configuration(self, extra):
        self.rparams.update(extra['rparams'])
        self.requested_walltime = extra['requested_walltime']
        extra_params = [
            self.rparams['memory'],
            self.rparams['fps'],
            self.rparams['pixel_to_scale'],
            self.rparams['difference_lag'],
            self.rparams['threshold1'],
            self.rparams['threshold2'],
        ]
        self.arguments[2:] = extra_params


class ParticleLocator(BemoviApp):
    application = 'plocator'

    def __init__(self, videofile, **extra):
        videofilename = os.path.basename(videofile)
        extra['jobname'] = "%s.%s" % (self.application, extra['videoname'])
        extra['output_dir'] = os.path.join(extra['base_output_dir'], self.application)

        scriptdir = os.path.dirname(__file__)

        self.ijoutbase = extra['videoname']+'.ijout.txt'
        outputfile = os.path.join('data',
                                  '2-particle',
                                  self.ijoutbase)
        rdatafile = os.path.join('data',
                                 '2-particle',
                                 'particle.RData')
        self.rdatafile = os.path.join(extra['output_dir'],
                                      'particle.RData')
        self.logfile = os.path.join(extra['output_dir'],
                                    'plocator.log')
        self.fps = extra['rparams']['fps']
        self.pixel_to_scale = extra['rparams']['pixel_to_scale']
        self.difference_lag = extra['rparams']['difference_lag']
        self.threshold1 = extra['rparams']['threshold1']
        self.threshold2 = extra['rparams']['threshold2']

        extra_params = [
            extra['rparams']['memory'],
            extra['rparams']['fps'],
            extra['rparams']['pixel_to_scale'],
            extra['rparams']['difference_lag'],
            extra['rparams']['threshold1'],
            extra['rparams']['threshold2'],
        ]
        gc3libs.Application.__init__(
            self,
            arguments = ["./plocator.sh", "locator"] + extra_params,
            inputs = {
                os.path.join(scriptdir, 'plocator.sh'): 'plocator.sh',
                os.path.join(scriptdir, 'bemovi.R'): 'bemovi.R',
                videofile: os.path.join('data', '1-raw', videofilename)},
            outputs = {outputfile : self.ijoutbase,
                       rdatafile: 'particle.RData'},
            stdout = 'plocator.log',
            join = True,
            **extra)


class ParticleLinker(BemoviApp):
    application = 'plinker'

    def __init__(self, particlefile, **extra):
        ijoutname = os.path.basename(particlefile)
        extra['jobname'] = "%s.%s" % (self.application, extra['videoname'])
        extra['output_dir'] = os.path.join(extra['base_output_dir'], self.application)
        scriptdir = os.path.dirname(__file__)
        rdatafile = os.path.join('data',
                                 '3-trajectory',
                                 'trajectory.RData')
        self.rdatafile = os.path.join(extra['output_dir'],
                                      'trajectory.RData')
        self.logfile = os.path.join(extra['output_dir'],
                                    'plinker.log')
        self.fps = extra['rparams']['fps']
        self.pixel_to_scale = extra['rparams']['pixel_to_scale']
        self.difference_lag = extra['rparams']['difference_lag']
        self.threshold1 = extra['rparams']['threshold1']
        self.threshold2 = extra['rparams']['threshold2']

        extra_params = [
            extra['rparams']['memory'],
            extra['rparams']['fps'],
            extra['rparams']['pixel_to_scale'],
            extra['rparams']['difference_lag'],
            extra['rparams']['threshold1'],
            extra['rparams']['threshold2'],
        ]
        gc3libs.Application.__init__(
            self,
            arguments = ["./plocator.sh", "linker"] + extra_params,
            inputs = {
                particlefile: os.path.join("data",
                                           "2-particle",
                                           os.path.basename(particlefile)),
                os.path.join(scriptdir, 'plocator.sh'): 'plocator.sh',
                os.path.join(scriptdir, 'bemovi.R'): 'bemovi.R',
            },
            outputs = {os.path.join('data',
                                    '3-trajectory',
                                    'ParticleLinker_' + ijoutname),
                       rdatafile},
            stdout = 'plinker.log',
            join = True,
            **extra)


class BemoviWorkflow(SequentialTaskCollection):
    appname = 'bemoviworkflow'
    def __init__(self, videofile, email_from, smtp_server, **extra):
        self.videofile = videofile
        videofilename = os.path.basename(videofile)

        inboxdir = os.path.dirname(videofile)
        self.vdescrfile = os.path.join(inboxdir, 'video.description.txt')
        self.inboxdir = inboxdir
        self.email_to = ''
        self.email_from = email_from
        self.smtp_server = smtp_server
        self.video_is_needed = True

        # Look for a 'configuration' file in the inboxdir to override
        # bemovi parameters

        videoname = videofilename.rsplit('.', 1)[0]

        self.extra = extra
        extra['jobname'] = 'BemoviWorkflow_%s' % videoname
        extra['videoname'] = videoname

        extra['base_output_dir'] = os.path.join(
            inboxdir + '.out',
            extra['jobname'])

        self.update_configuration()

        plocator = ParticleLocator(videofile, **extra)
        plinker = ParticleLinker(
            os.path.join(extra['base_output_dir'],
                         ParticleLocator.application,
                         extra['videoname'] + '.ijout.txt',),
            **extra)
        SequentialTaskCollection.__init__(self, [plocator, plinker], **extra)

    def update_configuration(self):
        csvcfgfile = os.path.join(self.inboxdir, 'gbemovi.csv')
        try:
            cfg = BemoviCSVConf(csvcfgfile).get(self.videofile)
            self.extra['rparams']['fps'] = cfg.get('fps')
            self.extra['rparams']['pixel_to_scale'] = cfg.get('pixel_to_scale')
            self.extra['rparams']['difference_lag'] = cfg.get('difference_lag')
            self.extra['rparams']['threshold1'] = cfg.get('threshold1')
            self.extra['rparams']['threshold2'] = cfg.get('threshold2')
            self.extra['requested_walltime'] = cfg.get('requested_walltime')
            self.video_is_needed = cfg.get('video_is_needed')
            self.email_to = cfg.get('email_to')
            for task in self.tasks:
                task.update_configuration(self.extra)
        except IOError:
            pass
        except Exception as ex:
            log.warning(
                "Error while reading CSV configuration file %s: Ignoring."
                " Error was: %s",
                csvcfgfile, ex)


    def should_resubmit(self):
        """Return True or False if the job should be resubmitted or not"""
        for task in self.tasks:
            if task.execution.state == Run.State.TERMINATED and \
               task.execution.exitcode != 0:
                return True
        return False

    def _send_notification(self, subject, body, attachments=None, debug=False):
        log.info("Sending notification to %s with subject \"%s\".", self.email_to, subject)
        if debug:
            log.debug("Subject: \"%s\"", subject)
            log.debug("Body: \"%s\"", body)
        else:
            try:
                msg = MIMEMultipart()
                msg['Subject'] = subject
                msg['From'] = self.email_from
                msg['To'] = self.email_to
                msg.attach(MIMEText(body))
                if attachments:
                    for attachment in attachments:
                        with open(attachment) as fd:
                            msg.attach(
                                MIMEApplication(fd.read(),
                                                name=attachment)
                            )
                s = smtplib.SMTP(self.smtp_server)
                s.sendmail(self.email_from, [self.email_to], msg.as_string())
                s.quit()
                log.info("Successfully sent email to %s", self.email_to)
            except Exception as ex:
                log.error("Error while sending an email to %s via %s: %s",
                          self.email_to, self.smtp_server, ex)


    def terminated(self):
        """Check if the processing went fine, otherwise send an email."""
        plocator, plinker = self.tasks
        if plocator.execution.exitcode != 0 or \
           plinker.execution.exitcode != 0:
            # Send notification

            # path of self.videofile should be relative to the inbox dir.
            shortvideofile = self.videofile.replace(os.path.dirname(self.inboxdir), '<inbox>')
            subject = "GBemovi: Failed processing file %s" % shortvideofile
            msg = """File {} failed processing.
ParticleLocator exited with status {}
ParticleLinker exited with status {}
""".format(shortvideofile,
           plocator.execution.exitcode,
           plinker.execution.exitcode)
            self._send_notification(
                subject,
                msg,
                attachments=[plocator.logfile,
                             plinker.logfile]
            )


class Merger(gc3libs.Application):
    appname = "merger"
    def __init__(self, vdescrfile, inputs, **extra):
        # Rename files in destination folder
        scriptdir = os.path.dirname(__file__)
        self.vdescrfile = vdescrfile
        self.inboxdir = os.path.dirname(vdescrfile)
        infiles = {
            os.path.join(scriptdir, 'plocator.sh'): 'plocator.sh',
            os.path.join(scriptdir, 'bemovi.R'): 'bemovi.R',
            vdescrfile: 'data/1-raw/video.description.txt',
        }
        for index, (ldata, tdata) in enumerate(inputs):
            infiles[ldata] = 'data/2-particle/particle-%05d.RData' % index
            infiles[tdata] = 'data/3-trajectory/trajectory-%05d.RData' % index
        extra['jobname'] = "Merger_%d_%s" % (index, time.strftime("%Y-%m-%d_%H.%M", time.localtime()))
        extra['output_dir'] = os.path.join(self.inboxdir + '.out', extra['jobname'])
        gc3libs.Application.__init__(
            self,
            arguments = ["./plocator.sh", "merger"],
            inputs = infiles,
            outputs = {"data/5-merged/Master.RData": "Master.RData"},
            # outputs = gc3libs.ANY_OUTPUT,
            stdout = 'merger.log',
            join = True,
            **extra)


class FinalMerger(Merger):
    """This application is exactly like Merger, but when it's terminated
    it will delete its input files"""
    appname = "final_merger"
    def __init__(self, vdescrfile, tasks, email_from, email_to, smtp_server, stats, session, **extra):
        self.email_from = email_from
        self.email_to = email_to
        self.smtp_server = smtp_server
        self.stats = stats
        self.videotasks = tasks
        self.videofiles = [t.videofile for t in tasks]
        self.session = session
        inputs = []
        for task in tasks:
            plinker, ptracker = task.tasks
            if plinker.execution.exitcode == ptracker.execution.exitcode == 0:
                inputs.append((plinker.rdatafile, ptracker.rdatafile))
        Merger.__init__(self, vdescrfile, inputs, **extra)

    def terminated(self):
        # if terminated with success, delete input files and send
        # notification
        if self.execution.exitcode != 0:
            # nothing to do
            del self.videotasks
            return
        try:
            text = """Report from experiment in inbox {}

Path to merged file: {}

Files processed:
{}

Statistics:
{}
""".format(os.path.dirname(self.vdescrfile),
           os.path.join(self.output_dir, 'merger.log'),
           str.join('\n', self.videofiles),
           self.stats)
            msg = MIMEText(text)
            msg['Subject'] = "GBemovi: Experiment in %s ended successfully." % os.path.dirname(self.vdescrfile)
            msg['From'] = self.email_from
            msg['To'] = str.join(', ', self.email_to)
            s = smtplib.SMTP(self.smtp_server)
            s.sendmail(self.email_from, self.email_to, msg.as_string())
            s.quit()
            log.info("Successfully sent email to %s", str.join(', ', self.email_to))
        except Exception as ex:
            log.error("Error while sending an email to %s via %s: %s",
                      self.email_to, self.smtp_server, ex)

        for videofile in self.videofiles:
            try:
                os.remove(videofile)
                log.info("Removed input file %s", videofile)
            except Exception as ex:
                log.warning("Ignorinig error while removing input file %s: %s",
                            videofile, ex)

        for task in self.videotasks:
            try:
                self.session.remove(task.persistent_id)
                task._controller.remove(task)
            except Exception as ex:
                log.warning("Ignoring error while removing task %s from session: %s", task.jobname, ex)
        # This is useful when the daemon is restarted...
        del self.videotasks

class GBemoviDaemon(SessionBasedDaemon):
    """Daemon to run bemovi workflow"""
    version = '1.0'

    def setup_options(self):
        self.add_param(
            '--fps',
            default='25',
            help="Framerate of the video in seconds (frame per second). Default: %(default)s")
        self.add_param(
            '--pixel-to-scale',
            default='1000/240',
            help="A conversion factor to scale videos (in pixel) to real"
            " dimensions (e.g.  micrometer). The conversion factor needs to"
            " be determined by the experimenter by measuring an object of"
            " known size (e.g. micrometer) with the used microscope/video"
            " settings (magnification, resolution etc.)."
            " Default: %(default)s")
        self.add_param(
            '--difference-lag',
            default='10',
            help="Lag between two frames, the former being the target frame"
            " to be segmented, the latter the frame subtracted from the former"
            " to create the difference image. Lag is specified in frames, but"
            " can be converted into time (difference-lag of 25 of a video"
            " taken with 25 fps translates into a difference-lag of 1 second)."
            " Default: %(default)s")
        self.add_param(
            '--threshold1',
            default='5',
            help="Threshold applied to the difference image. Threshold has to"
            " be in the range of 0 to 255. The lower the threshold, the more"
            " greyish pixels will be considered when the image is binarized."
            " Default: %(default)s")
        self.add_param(
            '--threshold2',
            default='255',
            help="Upper threshold applied to the difference image. Value of"
            " `--threshold2` has to be larger than `--threshold1` (lower"
            " threshold). Default: %(default)s")
        self.add_param(
            '--valid-extensions',
            default='avi,cxd,raw',
            help="Comma separated list of valid extensions for video file."
            " Files ending with an extension non listed here will be ignored."
            " Default: %(default)s")
        hostname = socket.gethostname()
        user = pwd.getpwuid(os.getuid()).pw_name
        email_from = '%s+gbemovi@%s' % (user, hostname)

        self.add_param(
            '--email-from',
            default=email_from,
            help="Email to use when sending notifications. Default: %(default)s")
        self.add_param(
            '--smtp-server',
            default='localhost',
            help="SMTP server to use to send notifications. Default: %(default)s")

    def setup_args(self):
        SessionBasedDaemon.setup_args(self)
        self.actions['notify_state'].default = "CLOSE_WRITE,CREATE"

    def parse_args(self):
        self.valid_extensions = [i.strip() for i in self.params.valid_extensions.split(',')]

        if not self.params.inbox:
            # Add a default inbox if not passed from command line
            self.params.inbox = [Url(os.path.join(self.params.working_dir, 'inbox'))]
        if self.params.output == self.actions['output'].default:
            # Use directory 'output' as output directory by default
            self.params.output = os.path.join(self.params.working_dir, 'output')

        if int(self.params.threshold2) <= int(self.params.threshold1):
            gc3libs.exceptions.InvalidUsage(
                "Value of `--threshold2` should be greater than `--threshold1`")
    def before_main_loop(self):
        # Setup new command
        self.comm.server.register_function(self.merge_data, "merge")
        self.comm.server.register_function(self.cleanup_session, "cleanup")

    def merge_data(self):
        # Scan all the output directories. Find the plocator.Rdata
        # files, add a merger for each input directory.
        scriptdir = os.path.dirname(__file__)

        # Walk through our jobs, group "experiments" and run a merger for each of them
        mergers = defaultdict(list)
        newapps = []
        for task in self.session:
            try:
                if task.appname != 'bemoviworkflow':
                    continue
            except AttributeError:
                # Some class might not have appname attribute
                continue
            if task.execution.state != Run.State.TERMINATED:
                # only merge completed jobs
                continue
            # FIXME: We should also avoid running the merger twice for
            # the same job

            # find plinker radata files
            plinker = task.tasks[0]
            ptracker = task.tasks[1]
            # If linker or tracker failed, skip the rdata file, as
            # it's likely to be empty and bemove will not like it.
            if plinker.execution.exitcode != 0 or \
               ptracker.execution.exitcode != 0:
                continue
            linkerdata = plinker.rdatafile
            trackerdata = ptracker.rdatafile
            mergers[task.vdescrfile].append((linkerdata, trackerdata))

        for vdescrfile, inputs in mergers.items():
            if not os.path.exists(vdescrfile):
                self.log.warning("Ignoring inbox directory %s as there is no video.description.txt file", os.path.dirname(vdescrfile))
                continue
            app = Merger(vdescrfile, inputs, **self.extra)
            newapps.append(app)
            self._controller.add(app)
            self.session.add(app)
            # FIXME: Submit merger application
            self.log.info("Creating Merger application from %d input videos" % len(inputs))

        msg = []
        for app in newapps:
            msg.append("Merger application created from directory %s" % app.inboxdir)
        if msg:
            return "Merging data for %d experiments.\n%s" % (len(msg), str.join('\n', msg))

    def cleanup_session(self):
        # for each experiment:
        #   => run "final" merger
        #   => delete input files
        #   => cleanup session
        #   => send report ("Experiment XXX final report")
        #
        # This is implemented via the FinalMerger application, which
        # is similar to the Merger application but is also responsible
        # for deleting the input files and sending a report after
        # termination.

        # group running jobs by their vdescrfile attribute
        experiments = defaultdict(list)
        not_completed = set()
        newapps = []
        for task in self.session:
            try:
                if task.appname != 'bemoviworkflow':
                    continue
            except AttributeError:
                continue
            if task.execution.state != Run.State.TERMINATED:
                continue
            experiments[task.vdescrfile].append(task)

        # Now we must filter non-completed experiments. There are two kinds:
        # 1) experiments with mandatory video files that failed
        # 2) experiments with files not yet processed.
        #
        # not_completed currently contains jobs of the first kind
        for vdescrfile, tasks in experiments.items():
            inboxdir = os.path.dirname(vdescrfile)
            try:
                bemovicfg = BemoviCSVConf(os.path.join(inboxdir, 'gbemovi.csv'))
            except Exception as ex:
                # Ignoring
                self.log.error(
                    "Unable to run FinalMerger application when gbemovi.csv is"
                    " missing or invalid. Ignoring directory %s: %s", inboxdir, ex)
                not_completed.add(vdescrfile)
                continue
            for task in tasks:
                if task.should_resubmit():
                    # We could access task.video_is_needed but the csv
                    # file might have been changed in the meantime.
                    if bemovicfg.get(task.videofile).get('video_is_needed'):
                        # Cannot merge this, experiment is not finished.
                        not_completed.add(task.vdescrfile)
                        continue
            if vdescrfile in not_completed:
                self.log.warning("Ignoring directory %s: some mandatory video were not correctly processed.", inboxdir)
                continue

            if os.path.exists(vdescrfile):
                # We need now to check if all the videos in the
                # experiment were processed
                videofiles = [os.path.basename(t.videofile) for t in tasks]
                fd = open(vdescrfile)
                data = fd.read().replace('\r', '\n')
                fd.close()
                # Note: we ignore the header line
                descriptions = [i.split()[0] for i in data.splitlines()[1:]]

                # Descriptions could miss the extension.
                intersection = set(descriptions).intersection(videofiles)
                fileext = ''
                if not intersection:
                    # description file are probably missing the extension.
                    for descr in descriptions:
                        for key in videofiles:
                            if key.startswith(descr):
                                fileext = key[len(descr):]
                                break
                        if fileext:
                            break
                descriptions = [i+fileext for i in descriptions]

                # Now we know the experiment is complete iff
                # descriptions is exactly the list of videofiles.
                if set(descriptions) != set(videofiles):
                    self.log.info("Inbox %s not yet completed",
                                  os.path.dirname(vdescrfile))
                    continue

                self.log.info("Creating FinalMerger for inbox %s",
                              os.path.dirname(vdescrfile))
                inputs = []
                email_to = set()
                stats = []
                for task in tasks:
                    plinker, ptracker = task.tasks
                    email_to.add(bemovicfg.get(task.videofile).get('email_to'))
                    stats.append("""Video file {}
  Linker
    duration: {}
    exitcode: {}

  Tracker
    duration: {}
    exitcode: {}
""".format(task.videofile,
           plinker.execution.duration, plinker.execution.exitcode,
           ptracker.execution.duration, ptracker.execution.exitcode))

                app = FinalMerger(vdescrfile,
                                  tasks,
                                  self.params.email_from,
                                  email_to,
                                  self.params.smtp_server,
                                  str.join('\n', stats),
                                  self.session,
                                  **self.extra)
                newapps.append(app)
                self._controller.add(app)
                self.session.add(app)
        if newapps:
            return "Running FinalMerger for the following inboxes:\n%s" % str.join('\n', [app.inboxdir for app in newapps])
        else:
            return "No experiment is done yet"

    def new_tasks(self, extra, epath=None, emask=0):
        extra['rparams'] = {
            'memory': str(self.params.memory_per_core.amount(unit=gc3libs.quantity.MB)),
            'fps': self.params.fps,
            'pixel_to_scale': self.params.pixel_to_scale,
            'difference_lag': self.params.difference_lag,
            'threshold1': self.params.threshold1,
            'threshold2': self.params.threshold2,
        }

        if not epath:
            # At startup, scan all the input directories and check if
            # there is a file which is not processed yet.

            # First, check which files we already did
            known_videos = {i.get('videofile', None): i for i in self.session}
            new_jobs = []
            for inboxurl in self.params.inbox:
                # WARNING: we assume this is a filesystem directory
                try:
                    inbox = inboxurl.path
                except:
                    inboxurl = Url(inboxurl)
                    inbox = inboxurl.path
                for dirpath, dirnames, fnames in os.walk(inbox):
                    for fname in fnames:
                        filename = os.path.join(dirpath, fname)
                        if filename.rsplit('.', 1)[-1] not in self.valid_extensions:
                            continue
                        if filename.startswith('._'):
                            self.log.warning("Ignoring file %s as it starts with '._'", filename)
                            continue
                        if filename not in known_videos:
                            app = BemoviWorkflow(filename,
                                                 self.params.email_from,
                                                 self.params.smtp_server,
                                                 **extra)
                            new_jobs.append(app)
                            known_videos[filename] = app
                        else:
                            # In case it exists, but the application
                            # termianted with an exit code, then we
                            # want to resubmit the job anyway
                            job = known_videos[filename]
                            if job.should_resubmit():
                                self.log.info("File %s might have been overwritten. Resubmitting job %s",
                                              filename, job.persistent_id)
                                # self._controller.kill(job)
                                # self._controller.progress()
                                # self._controller.redo(job, from_stage=0)
            return new_jobs

        fpath = epath.path
        if emask & plr.events['IN_CLOSE_WRITE']:
            if fpath.rsplit('.', 1)[-1] not in self.valid_extensions:
                self.log.info("Ignoring file %s as it does not end with a valid extension (%s)",
                              fpath, str.join(',', self.valid_extensions))
                return []

            # Only resubmit the job if it failed
            for job in self.session.tasks.values():
                try:
                    job.videofile
                except AttributeError:
                    # Not a GBemoviWorkflow application
                    continue
                if job.videofile == fpath:
                    if job.should_resubmit():
                        self.log.info("Re-submitting job %s as file %s has been overwritten",
                                      job.persistent_id, fpath)
                        job.update_configuration()
                        self._controller.kill(job)
                        self._controller.progress()
                        self._controller.redo(job, from_stage=0)

                    else:
                        self.log.info("Ignoring already successfully processed file %s", fpath)
                    # In both case, do not return any new job
                    return []
            return [BemoviWorkflow(fpath,
                                   self.params.email_from,
                                   self.params.smtp_server,
                                   **extra)]
        return []


if "__main__" == __name__:
    from gbemovi import GBemoviDaemon
    GBemoviDaemon().run()



#! /usr/bin/env python
#
#   gmphili.py -- Parallel runs of M. Th. Philipp's "stability" pipeline
#
#   Copyright (C) 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your option)
#   any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
"""

from __future__ import absolute_import, print_function
import csv
import os

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, existing_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gmphili
    gmphili.GMphiliScript().run()


__version__ = '1.0'
# summary of user-visible changes
__changelog__ = """
  2016-09-21:
  * Initial version
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


class GMphiliScript(SessionBasedScript):
    """
    Read a `.csv` file and invoke the 'stability' R code once per each row.

    The CSV input file is expected to have 5 columns: 'dgp', 'learner',
    'cldist', 'dim', 'n'. Each 'stability' job will be called with these 5
    arguments, in the order they are given in the input file.

    The ``gmphili`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gmphili`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version=__version__, # module version == script version
            application=GMphiliApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for=GMphiliApplication,
        )


    def setup_args(self):
        self.add_param(
            'scenarios', nargs='+', type=existing_file,
            help=(
                "CSV file(s) listing program invocation parameters."
                " It is expected to have the following 5 columns:"
                " 'dgp', 'learner', 'cldist', 'dim', 'n'."
            ))


    def new_tasks(self, extra):
        tasks = []
        for scenario in self.params.scenarios:
            with open(scenario, 'r') as input_file:
                header, dialect = _csv_features(input_file)
                input_csv = csv.reader(input_file, dialect)
                # skip header row
                if header:
                    input_csv.next()
                for row in input_csv:
                    extra_args = extra.copy()
                    self.log.info("Adding simulation with parameters %s", row)
                    tasks.append(GMphiliApplication(*row, **extra_args))

        return tasks


class GMphiliApplication(Application):
    """
    Run M. Philipp's `runsim.R` pipeline.
    """

    def __init__(self, dgp, learner, cldist, dim, n, **extra_args):
        # save execution parameters
        self.dgp = dgp
        self.learner = learner
        self.cldist = cldist
        self.dim = dim
        self.n = n
        # provide defaults
        extra_args.setdefault('requested_cores', 1)
        # initialize GC3Pie Application object
        super(GMphiliApplication, self).__init__(
            # command-line to run
            [
                # GC3Pie wrapper script
                './run_R.sh',
                # actual R code invocation
                'runsim', dgp, learner, cldist, dim, n,
                # how many cores to use for parallel code
                extra_args.get('requested_cores', 1),
            ],
            # input files
            inputs=[
                resource_filename(Requirement.parse('gc3pie'),
                                  'gc3libs/etc/run_R.sh'),
                'stability_v0.1-4.R',
                'runsim.R',
            ],
            # output files
            outputs=[
                'siminfo.rda',
                'simres.rda'
            ],
            stdout='runsim.log',
            stderr='runsim.log',
            # output dir, etc.
            **extra_args
        )


def _csv_features(input_file, sample_size=1024):
    """
    Guess dialect and other features of a CSV file from its initial contents.
    Return a tuple consisting of:

    * ``header``: a boolean flag indicating whether the first row is a header
      line;
    * ``dialect``: a `csv.Dialect` instance, or ``None`` if auto-detection of
      dialect failed.

    Reads the first `sample_size` bytes from the file and analyzes them, then
    rewinds the file to the beginning. Expects the passed file-like object to
    be seekable, and the stream to be positioned at the beginning of the CSV
    content.

    This is just a convenient interface for the `Sniffer`:class: in Python's
    standard `csv` module.
    """
    startpos = input_file.tell()
    detect = csv.Sniffer()
    sample = input_file.read(sample_size)
    dialect = detect.sniff(sample)
    header = detect.has_header(sample)
    input_file.seek(startpos, 0)  ## rewind
    return header, dialect



#!/usr/bin/env python

from __future__ import absolute_import, print_function
import numpy as n, pylab as p, time
import os

from pymods.support.support import wrapLogger

logger = wrapLogger(loggerName = 'convexHullLogger', streamVerb = 'INFO', logFile = os.path.join(os.getcwd(), 'convexHull.log'))


def _angle_to_points(points, centre):
    '''calculate angle in 2-D between points and x axis'''
    resOut = []
    for point in points:
        delta = point - centre
        res = n.arctan(delta[1] / delta[0])
        if delta[0] < 0:
            res += n.pi
        resOut.append(res)
    return n.array(resOut)

def _angle_to_point(point, centre):
    '''calculate angle in 2-D between points and x axis'''
    delta = point - centre
    res = n.arctan(delta[1] / delta[0])
    if delta[0] < 0:
        res += n.pi
    return res


def _draw_triangle(p1, p2, p3, **kwargs):
    tmp = n.vstack((p1,p2,p3))
    x,y = [x[0] for x in zip(tmp.transpose())]
    p.fill(x,y, **kwargs)
    #time.sleep(0.2)


def area_of_triangle(p1, p2, p3):
    '''calculate area of any triangle given co-ordinates of the corners'''
    return n.linalg.norm(n.cross((p2 - p1), (p3 - p1)))/2.


def convex_hull(points, graphic=True, smidgen=0.0075):
    '''Calculate subset of points that make a convex hull around points

    Recursively eliminates points that lie inside two neighbouring points until only convex hull is remaining.
    
    :Parameters:
        points : ndarray (2 x m) 
            array of points for which to find hull
        graphic : bool
            use pylab to show progress?
        smidgen : float
            offset for graphic number labels - useful values depend on your data range

    :Returns:
        hull_points : ndarray (2 x n)
            convex hull surrounding points
    '''
    logger.debug('start convex hull')

    if graphic:
        p.clf()
        p.plot(points[0], points[1], 'ro')
    n_pts = points.shape[1]
    assert(n_pts > 1)
    centre = points.mean(1)
    if graphic: p.plot((centre[0],),(centre[1],),'bo')
    angles = n.apply_along_axis(_angle_to_point, 0, points, centre)

    logger.debug('angles: %s' %angles)

    pts_ord = points[:,angles.argsort()]
    logger.debug('pts_ord: %s' %pts_ord)

    if graphic:
        for i in xrange(n_pts):
            p.text(pts_ord[0,i] + smidgen, pts_ord[1,i] + smidgen, '%d' % i)
    pts = [x[0] for x in zip(pts_ord.transpose())]
    logger.debug('pts: %s' %pts)
    prev_pts = len(pts) + 1
    k = 0
    while prev_pts > n_pts:
        logger.debug(' ')
        logger.debug('new round of points')
        logger.debug('current # of points %d' %n_pts )
        prev_pts = n_pts
        n_pts = len(pts)
        if graphic: p.gca().patches = []
        i = -2
        while i < (n_pts - 2):
            logger.debug('i: %d' % i)
            Aij = area_of_triangle(centre, pts[i],     pts[(i + 1) % n_pts])
            logger.debug('Aij: %s' % Aij)
            Ajk = area_of_triangle(centre, pts[(i + 1) % n_pts], pts[(i + 2) % n_pts])
            logger.debug('Aik: %s' % Ajk)
            Aik = area_of_triangle(centre, pts[i],     pts[(i + 2) % n_pts])
            logger.debug('Aik: %s' % Aik)
            if graphic:
                _draw_triangle(centre, pts[i], pts[(i + 1) % n_pts], facecolor='blue', alpha = 0.2)
                _draw_triangle(centre, pts[(i + 1) % n_pts], pts[(i + 2) % n_pts], facecolor='green', alpha = 0.2)
                _draw_triangle(centre, pts[i], pts[(i + 2) % n_pts], facecolor='red', alpha = 0.2)
            if Aij + Ajk < Aik:
                if graphic: 
                    p.plot((pts[i + 1][0],),(pts[i + 1][1],),'go')
                del pts[i+1]
            i += 1
            n_pts = len(pts)
        k += 1
    return n.asarray(pts).transpose()


def project_point_to_line_segment(A, B, points):
    '''
    Parameters:
      A: Vector of beginning line points
      B: Vector of ending line points
      points: The points for which the projection is to be done. 
    Returns:
      qOut: the closest points (projections) on the line segment
    '''
    if A.ndim == 1:
        A = n.array([A])
        B = n.array([B])
        points = n.array([points])

    qOut = n.empty( (0, 2) )
    for a, b, point in zip(A,B, points):
        # vector from A to B
        AB = (b-a)
        # squared distance from A to B
        AB_squared = n.dot(AB,AB)

        if AB_squared == 0:
            # A and B are the same point
            q = a
        else:
            # vector from A to p
            Ap = (point-a)
            # from http://stackoverflow.com/questions/849211/
            # Consider the line extending the segment, parameterized as A + t (B - A)
            # We find projection of point p onto the line.
            # It falls where t = [(p-A) . (B-A)] / |B-A|^2
            t = n.dot(Ap,AB)/AB_squared
            if t < 0.0:
            # "Before" A on the line, just return A
                q = a
            elif t > 1.0:
                # "After" B on the line, just return B
                q = b
            else:
                # projection lines "inbetween" A and B on the line
                q = a + t * AB
            #q = q[:, n.newaxis]
            q = n.array([q])
        qOut = n.append(qOut, q, 0)
    return qOut


def lineFromTwoPoints(point1, point2):
    ''' creates line, specified as: a * x + b * y = c '''
    x1 = point1[0]
    y1 = point1[1]
    x2 = point2[0]
    y2 = point2[1]
    # a = point2[1] - point1[1]
    # b = point1[0] - point2[0]
    # c = a * point1[0] + b * point1[0]
    a = y2 - y1
    b = x1 - x2
    c = a * x1 + b * y1
    
    return n.array([a,b,c])
    

def lineLineIntersection(line1, line2):
    ''' from http://community.topcoder.com/tc?module=Static&d1=tutorials&d2=geometry2'''
    A = n.array([line1[0], line2[0]])
    B = n.array([line1[1], line2[1]])
    C = n.array([line1[2], line2[2]])
                
    det = A[0] * B[1] - A[1] * B[0]
    if (det == 0):
        return False
    else:
        x = (B[1] * C[0] - B[0] * C[1]) / det
        y = (A[0] * C[1] - A[1] * C[0]) / det
        return n.array([x,y])
        

if __name__ == "__main__":
    print('entered main')
    #points = n.random.random_sample((2,40))
    #print(points)
    #hull_pts = convex_hull(points)

#    p.show()
    # p.clf()
    # p.plot(points[0], points[1], 'ro')


#    pylab.savefig('simple_plot3') # 
#    p.show() 

    # print(hull_pts)

    A = n.array([[1, 2], [2, 4]])
    B = n.array([[3, 4], [1, 2]])
    p = n.array([[3.5, 10], [3,4]])
    q = project_point_to_line_segment(A, B, p)
    print(q)
    
    _angle_to_point(n.array([[1, 4], [1, 6]]), n.array([2, 4]))
    
    print('done')






#! /usr/bin/env python
#
#   gperm.py -- Front-end script for running Docker BIDS apps
#   function over a large dataset.
#
#   Copyright (C) 2017, 2018  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `BIDS apps` jobs fetching
docker images from the `BIDS apps` repository.

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gperm.py --help`` for program usage
instructions.

Example of docker execution:
docker run -i --rm -v /mnt/filo/data/ds005:/bids_dataset:ro -v /mnt/filo/outputs:/outputs bids/fmriprep /bids_dataset /outputs participant --participant_label 01

gperm takes BIDS files as input.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2018-03-10:
  * result of each subject get stored back to the corresponding input folder.
  2018-01-10:
  * added support for freesurfer license file to be passed as part of the docker invokation. see: https://fmriprep.readthedocs.io/en/latest/installation.html#the-freesurfer-license
  2017-04-18:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gperm
    gperm.GpermScript().run()

import os
import sys
import time
import tempfile
import mimetypes
import random
import shutil

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, positive_int, existing_file, existing_directory
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

PHOENY_OUTPUT_DIR="output"
DEFAULT_BIDS_FOLDER="data/"
DEFAULT_RESULT_FOLDER="output/"
DEFAULT_DOCKER_BIDS_ARGS="--no-submm-recon"
DEFAULT_FREESURFER_LICENSE_FILE="license.txt"
DEFAULT_DOCKER_BIDS_APP="poldracklab/fmriprep "+DEFAULT_DOCKER_BIDS_ARGS
DEFAULT_REPETITIONS=1

DOCKER_RUN_COMMAND="sudo docker run -i --rm {DOCKER_MOUNT} {DOCKER_TO_RUN} /bids /output participant --participant_label {SUBJECT_NAME}"

## Utility methods
def _get_subjects(input_folder):
    """
    return tuple ([list .json and .tsv files],[list of sub-folders])
    Assumptions:
    * each sub-folder contains a valid subject's data
    * each .json and .tsv file found in root folder will be made available
    to all Applications.
    """
    control_files_list = []
    subjects_folders_list = []

    for element in os.listdir(input_folder):
        full_element = os.path.abspath(os.path.join(input_folder,element))
        if element.endswith(".json") or element.endswith(".tsv"):
            # Valid control file
            control_files_list.append(full_element)

        elif os.path.isdir(full_element):
            # Valid subject folder
            subjects_folders_list.append(full_element)

    return (control_files_list,subjects_folders_list)

## custom application class
class GpermApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gperm'

    def __init__(self, subject, subject_name, control_files, repeat_index, docker_run, freesurfer_license, **extra_args):

        executables = []
        inputs = dict()
        outputs = dict()

        self.subject_dir = subject
        inputs[subject] = os.path.join(DEFAULT_BIDS_FOLDER,
                                       os.path.basename(subject))

        # add all control files to 'data' folder
        for element in control_files:
            inputs[element] = os.path.join(DEFAULT_BIDS_FOLDER,
                                       os.path.basename(element))

        if not os.path.isdir(DEFAULT_RESULT_FOLDER):
            os.mkdir(DEFAULT_RESULT_FOLDER)
        inputs[DEFAULT_RESULT_FOLDER] = DEFAULT_RESULT_FOLDER

        # Define mount points
        DOCKER_MOUNT=" -v $PWD/{SUBJECT_DIR}:/bids:ro -v $PWD/{OUTPUT_DIR}:/output ".format(SUBJECT_DIR=DEFAULT_BIDS_FOLDER,
                                                                                     OUTPUT_DIR=DEFAULT_RESULT_FOLDER)

        if freesurfer_license:
            inputs[freesurfer_license] = os.path.basename(freesurfer_license)
            DOCKER_MOUNT+=" -v $PWD/{0}:/opt/freesurfer/license.txt ".format(inputs[freesurfer_license])

        arguments = DOCKER_RUN_COMMAND.format(DOCKER_MOUNT=DOCKER_MOUNT,
                                              DOCKER_TO_RUN=docker_run,
                                              SUBJECT_NAME=subject_name)

        gc3libs.log.debug("Creating application for executing: %s", arguments)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = [DEFAULT_RESULT_FOLDER],
            stdout = 'gperm.log',
            join=True,
            executables = executables,
            **extra_args)

    def terminated(self):
        """
        Move output file in 'result_dir'
        """
        gc3libs.log.info("Moving results from {0} to {1}".format(self.output_dir,
                                                                 self.subject_dir))

        for elem in os.listdir(self.output_dir):
            shutil.move(os.path.join(self.output_dir,elem),
                        os.path.join(self.subject_dir,elem))


class GpermScript(SessionBasedScript):
    """
    The ``gperm`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gperm`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gperm``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GpermApplication,
            stats_only_for = GpermApplication,
            )

    def setup_args(self):
        self.add_param('input_folder',
                       type=existing_directory,
                       help="Path to input folder containing valid input .bids files.")

    def setup_options(self):
        self.add_param("-R", "--repeat", metavar="[INT]",
                       type=positive_int,
                       dest="repeat", default=DEFAULT_REPETITIONS,
                       help="Repeat analysis. Default: %(default)s.")

        self.add_param("-L", "--license", metavar="[PATH]",
                       type=existing_file,
                       dest="freesurfer_license", default=None,
                       help="Location of freesurfer license file. Default: %(default)s.")

        self.add_param("-D", "--docker", metavar="[PATH]",
                       dest="docker", default=DEFAULT_DOCKER_BIDS_APP,
                       help="BIDS app docker image and execution arguments. Default: '%(default)s'.")


    def parse_args(self):
        """
        parse 'docker' command and separate docker image reference
        from execution arguments
        """

        self.docker_image = self.params.docker.split(' ')[0]
        self.docker_args =  self.params.docker.split(' ')[1:]

    def new_tasks(self, extra):
        """
        For each valid input file create a new GpermRetryableTask
        """
        tasks = []

        control_files,subjects_list = _get_subjects(self.params.input_folder)
        for subject in subjects_list:
            for repeat in range(0,self.params.repeat):
                subject_name = os.path.basename(subject)
                jobname = "{subject}-{rep}".format(subject=subject_name,
                                                   rep=repeat)

                extra_args = extra.copy()

                extra_args['jobname'] = jobname
                extra_args['output_dir'] = self.params.output
                extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

                self.log.debug("Creating Application for subject %s repetition : %d" % (subject_name,
                                                                                        repeat))

                tasks.append(GpermApplication(
                    subject,
                    subject_name,
                    control_files,
                    repeat,
                    self.params.docker,
                    self.params.freesurfer_license,
                    **extra_args))

        return tasks



#!/usr/bin/env python

from __future__ import absolute_import, print_function
import os
import numpy as np
import scipy.optimize
import scipy.interpolate as si
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
import pylab as p

from pymods.support.support import wrapLogger
import convexHull


logger = wrapLogger(loggerName = 'iwdInterpolationLogger', streamVerb = 'DEBUG', logFile = os.path.join(os.getcwd(), 'iwdInterpolation.log'))

np.seterr(all='raise')

class iwdInterpolation(object):
    '''
    Inverse weighted density interpolation
    '''
    def __init__(self, xMat, fx, normExp = 4, dx = 1e-3, makePlot=False):
        logger.debug('initializing new instance of iwdInterpolation')
        self.center = xMat.mean(0)
        angles = convexHull._angle_to_points(xMat, self.center)
        sortIndices = angles.argsort()
        self.angles = angles[sortIndices]
        self.xMat = xMat[sortIndices, :]
        self.fx = fx[sortIndices]
        self.normExp = normExp
        self.dx = dx
        self.makePlot = makePlot

        self.hullMat = convexHull.convex_hull(self.xMat.T, False).T
        self.hullAngles = convexHull._angle_to_points(self.hullMat, self.center)
        self.hullPoints = len(self.hullMat)

        if self.makePlot:
            self.plot()
        
    def __call__(self, x0):        
        if np.array(x0).ndim == 1:
            x0 = np.array([x0])
        gxList = []
        for x0Ele in x0:
            x0Ele = np.array(x0Ele)
            isInside = self.check_is_inside(x0Ele)
            logger.debug('isInside: %s' % isInside)
            if isInside:
                if x0Ele in self.xMat:
                    gx = self.fx[ np.all(x0Ele == self.xMat, 1) ]
                else:
                    dVec = self.d(x0Ele, self.xMat)
                    gx = np.sum( dVec * self.fx ) / np.sum(dVec)
                gxList.append(gx)
            else:
                hullPointPrev, hullPointNext = self.closest_hull_points(x0Ele)
                
                projected_x0 = convexHull.project_point_to_line_segment(hullPointPrev, hullPointNext, x0Ele)[0]
                logger.debug('projected_x0: %s' % projected_x0)
                
                if self.makePlot:
                    p.plot(x0Ele[0], x0Ele[1], 'ko')
                    p.plot(projected_x0[0], projected_x0[1], 'ro')
                
                q1 = projected_x0 - 1e-14 * (x0Ele - projected_x0) #be safely inside
                
                exitPoint = self.findExit(x0Ele,q1)
                
                if self.makePlot:
                    p.plot(exitPoint[0], exitPoint[1], 'go')

                logger.debug('q1: %s' % q1)
                # logger.debug('x0Ele - q1 %s' % (x0Ele - q1))
                # logger.debug('norm %s' % np.linalg.norm(x0Ele - q1) )
#                q2 = q1 - self.dx * (x0Ele - q1) / np.linalg.norm(x0Ele - q1) 
                q2 = (q1 + exitPoint) / 2
                logger.debug('q2: %s' % q2)
                
                f_q1 = self.__call__(q1)
                f_q2 = self.__call__(q2)
                logger.debug('f_q1: %s' % f_q1)
                logger.debug('f_q2: %s' % f_q2)
                df = (f_q1 - f_q2) / np.linalg.norm(q1 - q2)
                logger.debug('df: %s' % df)
                fx0 = f_q1 + np.linalg.norm(x0Ele - q1) * df
                logger.debug('fx0: %s' % fx0)
                gxList.append(fx0[0])
        return np.array(gxList)     

    def findExit(self,x0,q1):
        logger.debug('entered find exit')
        lineX0 = convexHull.lineFromTwoPoints(x0,q1)
        logger.debug('lineX0: %s' % lineX0)

        for pos in range(0, self.hullPoints):
            logger.debug('pos: %s' % pos)
            if pos == 0:
                lineHull = convexHull.lineFromTwoPoints(self.hullMat[self.hullPoints-1], self.hullMat[pos])
            else:
                lineHull = convexHull.lineFromTwoPoints(self.hullMat[pos-1], self.hullMat[pos])
            exitPoint = convexHull.lineLineIntersection(lineX0, lineHull)
            logger.debug('exitPoint: %s' % exitPoint)
#            p.plot(exitPoint[0], exitPoint[1], 'ro')
            
            # p.plot(self.hullMat[self.hullPoints-1,0],self.hullMat[self.hullPoints-1,0],'go')
            # p.plot(self.hullMat[pos,0],self.hullMat[pos,1],'go')
            # p.show()
            # exit()
            if (np.linalg.norm(exitPoint - q1) > 1e-10):
                if (self.check_is_inside(exitPoint)):
                    return exitPoint
        print('trouble --- no exit found')
        exit()
            


    def closest_hull_points(self, x0Ele):
        ''' finds the previous and next point on the hull'''
        angle_x0 = convexHull._angle_to_point(x0Ele, self.center)        
        logger.debug('angle_x0: %s' % angle_x0)
        pos = [i for i,x in enumerate(angle_x0 > self.hullAngles) if x == False]
        if len(pos) == 0:
            pos = 0
        else:
            pos = pos[0]
    
        if pos >= 0:
            hullPointPrev = self.hullMat[pos-1]
        else:
            hullPointPrev = self.hullMat[self.hullPoints-1]
        logger.debug('hullPointPrev: %s' % hullPointPrev)
        hullPointNext = self.hullMat[pos]
        logger.debug('hullPointNext: %s' % hullPointNext)
        return hullPointPrev, hullPointNext

    def d(self, x1, x2):
        distanceVectors = x1 - x2
        distanceNorms = np.array([ np.linalg.norm(distanceVector) for distanceVector in distanceVectors ])
        d = 1.0 / distanceNorms**self.normExp
        return d

    def plot(self):
        p.clf()
        
        p.plot(self.xMat[:, 0], self.xMat[:, 1], 'rx')
        p.plot(self.hullMat[:, 0], self.hullMat[:, 1], 'g+')
        p.plot((self.center[0],),(self.center[1],),'bo')
        
#        p.show()
    
    def check_is_inside(self, x0):
        logger.debug('entering is inside')
        logger.debug('x0: %s' %x0)

        hullPointPrev, hullPointNext = self.closest_hull_points(x0)
        A_prev_next = convexHull.area_of_triangle(self.center, hullPointPrev, hullPointNext)
        A_prev_x0 = convexHull.area_of_triangle(self.center, hullPointPrev, x0)
        A_next_x0 = convexHull.area_of_triangle(self.center, hullPointNext, x0)

        if A_prev_next + 1e-4 > A_prev_x0 + A_next_x0:
            isInside = True
        else:
            isInside = False
        logger.debug('exiting is inside')
        return isInside
        


if __name__ == '__main__':   
    np.random.seed(123)
    nPoints = 5
    makePlot = True
    normExp = 4
    dx = 1e-7
    
    def fun(x):
        x = np.asarray(x)
        if x.ndim == 1:
            x = np.array([x])
        return np.array([ np.sum(np.log(xEle+1)) for xEle in x ])
#        return np.array([ np.sum(xEle) for xEle in x ])
        

    xMat = np.random.random_sample((nPoints, 2))
    fx = fun(xMat)

    print(fx )

    iwd = iwdInterpolation(xMat, fx, normExp, dx, makePlot)

    print('evaluation with point outside:')
    testVec = np.array([1,0.6])
    print(iwd.check_is_inside(testVec))
    print(iwd(testVec))
 
    print('exact value' )
    print(fun(testVec))

    if makePlot:
        p.show()

    exit()

    print('evaluation with point inside:')
    testVec = np.array([0.5, 0.6])
    print(iwd.check_is_inside(testVec))
    print(iwd(testVec))

    print('exact value' )
    print(fun(testVec))

    print('done')




# test_workflow.py
# -*- coding: utf-8 -*-
#
#  Copyright (C) 2015, 2016  University of Zurich. All rights reserved.
#
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# stdlib imports
from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from io import StringIO
import os
import shutil
import tempfile
import re

# GC3Pie imports
from gc3libs import Run
from gc3libs.workflow import SequentialTaskCollection, StagedTaskCollection, StopOnError

from gc3libs.testing.helpers import SuccessfulApp, UnsuccessfulApp, temporary_core


def test_staged_task_collection_progress():
    class ThreeStageCollection(StagedTaskCollection):
        def stage0(self):
            return SuccessfulApp()
        def stage1(self):
            return SuccessfulApp()
        def stage2(self):
            return UnsuccessfulApp()

    with temporary_core() as core:
        coll = ThreeStageCollection()
        coll.attach(core)
        coll.submit()
        assert coll.execution.state == Run.State.SUBMITTED

        # first task is successful
        while coll.tasks[0].execution.state != Run.State.TERMINATED:
            coll.progress()
        assert coll.execution.state in [Run.State.SUBMITTED, Run.State.RUNNING]
        #assert_equal(coll.execution.exitcode, 0)

        # second task is successful
        while coll.tasks[1].execution.state != Run.State.TERMINATED:
            coll.progress()
        #assert_equal(coll.execution.state, Run.State.RUNNING)
        #assert_equal(coll.execution.exitcode, 0)

        # third task is unsuccessful
        while coll.tasks[2].execution.state != Run.State.TERMINATED:
            coll.progress()
        assert coll.execution.state == Run.State.TERMINATED
        assert coll.execution.exitcode == 1


def test_staged_task_collection_stage():
    class TwoStageCollection(StagedTaskCollection):
        def stage0(self):
            return SuccessfulApp(name='stage0')
        def stage1(self):
            return UnsuccessfulApp(name='stage1')

    with temporary_core() as core:
        coll = TwoStageCollection()
        coll.attach(core)
        coll.submit()
        stage = coll.stage()
        assert isinstance(stage, SuccessfulApp), ("stage=%r" % (stage,))
        assert stage.jobname == 'stage0'

        # advance to next task
        while coll.tasks[0].execution.state != Run.State.TERMINATED:
            coll.progress()
        stage = coll.stage()
        assert isinstance(stage, UnsuccessfulApp)
        assert stage.jobname == 'stage1'



#! /usr/bin/env python

"""
session - persistent collections of tasks
"""

# Copyright (C) 2012-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
__docformat__ = 'reStructuredText'


# stdlib imports
import atexit
import csv
import errno
import os
import sys
import shutil
import tempfile

# GC3Pie imports
import gc3libs
import gc3libs.exceptions
import gc3libs.persistence
import gc3libs.persistence.store
import gc3libs.utils
from gc3libs.utils import to_str
from gc3libs.workflow import TaskCollection


class Session(list):
    """
    A 'session' is a persistent collection of tasks.

    Tasks added to the session are persistently recorded using an
    instance of `gc3libs.persistence.Store`.  Stores can be shared
    among different sessions: each session knows wich jobs it 'owns'.

    A session is associated to a directory, which holds all the data
    releated to that session. Specifically, two files are always
    created in the session directory andused internally by this class:

    * `index.txt`: contains a list of all job IDs
      associated with this session;
    * `store.url`:  its contents are the URL of the store to create
      (as would be passed to the `gc3libs.persistence.make_store` factory).

    To only argument needed to instantiate a session is the `path` of
    the directory; the directory name will be used as the identifier
    of the session itself.  For example, the following code creates a
    temporary directory and the two files mentioned above inside it::

        >>> import tempfile; tmpdir = tempfile.mktemp(dir='.')
        >>> session = Session(tmpdir)
        >>> for name in sorted(os.listdir(tmpdir)):
        ...   print(name)
        created
        session_ids.txt
        store.url

    When a `Session` object is created with a `path` argument pointing
    to an existing valid session, the index of jobs is automatically
    loaded into memory, and the store pointed to by the ``store.url``
    file in the session directory will be used, *disregarding the
    contents of the `store_url` argument.*

    In other words, the `store_url` argument is only used when
    *creating a new session*.  If no `store_url` argument is passed
    (i.e., it has its default value), a `Session` object will
    instantiate and use a `FileSystemStore`:class: store, keeping data
    in the ``jobs`` subdirectory of the session directory.

    Methods `add` and `remove` are provided to manage the collection;
    the `len()` operator returns the number of tasks in the session;
    iteration over a session returns the tasks one by one::

        >>> task1 = gc3libs.Task()
        >>> id1 = session.add(task1)
        >>> task2 = gc3libs.Task()
        >>> id2 = session.add(task2)
        >>> len(session)
        2
        >>> for t in session:
        ...     print(type(t))
        <class 'gc3libs.Task'>
        <class 'gc3libs.Task'>
        >>> session.remove(id1)
        >>> len(session)
        1

    When passed the `flush=False` optional argument, methods `add` and
    `remove` do not update the session metadata: i.e., the tasks are
    added or removed from the store and the in-memory task list, but
    the updated task list is not saved back to disk.  This is useful
    when making many changes in a row; call `Session.flush` to persist
    the full set of changes.

    The `Store`:class: object is anyway accessible in the
    `store`:attr: attribute of each `Session` instance::

        >>> type(session.store)
        <class 'gc3libs.persistence.filesystem.FilesystemStore'>

    However, `Session` defines methods `save` and `load` as a
    convenient proxy to the corresponding `Store` methods::

        >>> obj = gc3libs.persistence.Persistable()
        >>> oid = session.save(obj)
        >>> obj2 = session.load(oid)
        >>> obj.persistent_id == obj2.persistent_id
        True

    The whole session data can be removed by using method `destroy`::

        >>> session.destroy()
        >>> os.path.exists(session.path)
        False

    """

    INDEX_FILENAME = 'session_ids.txt'
    STORE_URL_FILENAME = 'store.url'
    TIMESTAMP_FILES = {
        'start': 'created',
        'end':   'finished'
    }

    DEFAULT_JOBS_DIR = 'jobs'

    def __init__(self, path, create=True, store_or_url=None,
                 load=True, task_ids=None, **extra_args):
        """
        First argument `path` is the path to the session directory.

        The `create` argument is used to control the behavior in case the
        session directory does not exists already:

        * If `create` is ``True`` (default) and the session directory
          does not exists then a new session will be created.

        * If `create` is ``False`` and the session directory does not
          exists an error will be raised.

        Optional argument `store_or_url` is *either* an existing valid
        `gc3libs.persistence.store.Store` instance, *or* the URL of
        the store, as would be passed to function
        `gc3libs.persistence.store.make_store`:func:; any additional
        keyword arguments are passed to `make_store` unchanged.

        .. note::

          The optional `store_or_url` argument and following keyword arguments
          are used if and only if a new session is being *created*; they are
          ignored when loading existing sessions!

        By default `gc3libs.persistence.filesystem.FileSystemStore`:class:
        (which see) is used for providing a new session with a store.

        Finally, if optional argument `load` is ``False`` then an
        already-existing session at `path` will be discarded and a new
        one will be created in its place. By default, `load` is
        ``True``, meaning that data from existing sessions is loaded
        into memory.
        """
        self.path = os.path.abspath(to_str(path, 'filesystem'))
        self.name = os.path.basename(self.path)
        self.tasks = {}
        # Session not yet created
        self.created = -1
        self.finished = -1

        # load or make session
        if os.path.isdir(self.path) and load:
            # Session already exists?
            try:
                self._load_session(task_ids=task_ids, **extra_args)
            except IOError as err:
                gc3libs.log.debug("Cannot load session '%s': %s", path, err)
                # "No such file or directory" => create new session
                if err.errno == errno.ENOENT:
                    # only create session if no task IDs were given
                    if create and not task_ids:
                        gc3libs.log.debug(
                            "Assuming session is incomplete or corrupted,"
                            " creating it again.")
                        self._create_session(store_or_url, **extra_args)
                    else:
                        raise gc3libs.exceptions.InvalidArgument(
                            "Directory '%s' does not contain a valid session" %
                            self.path)
                else:
                    raise
        else:
            assert not task_ids, (
                "Creating a new session with non-empty `task_ids` makes no sense!")
            if create:
                self._create_session(store_or_url, **extra_args)
            else:
                raise gc3libs.exceptions.InvalidArgument(
                    "Session directory '%s' not found" % self.path)

    def _create_session(self, store_or_url, **extra_args):
        if isinstance(store_or_url, gc3libs.persistence.store.Store):
            self.store = store_or_url
        else:
            if store_or_url is None:
                store_or_url = os.path.join(self.path, self.DEFAULT_JOBS_DIR)
            # Ensure session directory exists before `make_store` is
            # called, or else SQLite raises an "OperationalError:
            # unable to open database file None None"
            gc3libs.utils.mkdir(self.path)
            self.store = gc3libs.persistence.make_store(store_or_url, **extra_args)
        self.store_url = self.store.url
        self._save_store_url_file()
        self._save_index_file()

        # Save the current command line on ``created`` file.
        with open(os.path.join(
                self.path,
                self.TIMESTAMP_FILES['start']), 'w') as fd:
            fd.write(' '.join(sys.argv) + '\n')

        self.set_start_timestamp()

    def _load_session(self, store=None, task_ids=None, **extra_args):
        """
        Load an existing session from disk.

        If a `store` argument is given, it is used as the session's
        task store; otherwise a new task store is created, passing
        keyword arguments to the `make_store` factory method.

        Any error that occurs while loading jobs from disk is ignored,
        unless environment variable ``GC3PIE_NO_CATCH_ERRORS``
        includes the keywords ``session`` or ``persistence``.
        """
        if store is not None:
            self.store = store
            self.store_url = self.store.url
        else:
            store_filename = os.path.join(self.path, self.STORE_URL_FILENAME)
            try:
                self.store_url = \
                    gc3libs.utils.read_contents(store_filename).strip()
                gc3libs.log.debug(
                    "Loading session from URL %s ...", self.store_url)
            except (OSError, IOError) as err:
                gc3libs.log.error(
                    "Unable to load session tasks from file `%s`: %s",
                    store_filename, err)
                raise
            self.store = gc3libs.persistence.make_store(
                self.store_url, **extra_args)

        if task_ids is None:
            idx_filename = os.path.join(self.path, self.INDEX_FILENAME)
            try:
                with open(idx_filename) as idx_file:
                    task_ids = idx_file.read().split()
            except (OSError, IOError) as err:
                gc3libs.log.error(
                    "Unable to load session index from file `%s`: %s",
                    idx_filename, err)
                raise

        try:
            start_file = os.path.join(
                self.path, self.TIMESTAMP_FILES['start'])
            self.created = os.stat(start_file).st_mtime
        except OSError:
            # FIXME: should we also set `self.created` to the current timestamp?
            gc3libs.log.warning(
                "Unable to recover starting time from existing session:"
                " file `%s` is missing.", start_file)

        self.tasks = self.load_many(task_ids, flush=False)

    def destroy(self):
        """
        Remove the session directory and all the tasks it contains
        from the store which are associated to this session.

        .. note::

          This will remove the associated task storage *if and only if*
          the storage is contained in the session directory!

        """
        for task_id in self.tasks:
            self._recursive_remove_from_store(task_id)
        if os.path.exists(self.path):
            shutil.rmtree(self.path)

    #
    # collection management
    #

    def add(self, task, flush=True):
        """
        Add a `Task` to the current session, save it to the associated
        persistent storage, and return the assigned `persistent_id`::

            >>> # create new, empty session
            >>> import tempfile; tmpdir = tempfile.mktemp(dir='.')
            >>> session = Session(tmpdir)
            >>> len(session)
            0

            >>> # add a task to it
            >>> task = gc3libs.Task()
            >>> tid1 = session.add(task)
            >>> len(session)
            1

        Duplicates are silently ignored: the same object can be added
        many times to the session, but gets the same ID each time::

            >>> # add a different task
            >>> tid2 = session.add(task)
            >>> len(session)
            1
            >>> tid1 == tid2
            True

            >>> # do cleanup
            >>> session.destroy()
            >>> os.path.exists(session.path)
            False

        """
        try:
            task_id = task.persistent_id
        except AttributeError:
            task_id = self.store.save(task)
        self.tasks[task_id] = task
        if flush:
            self.flush()
        return task_id

    def forget(self, task_id, flush=True):
        """
        Remove task identified by `task_id` from the current session
        *but not* from the associated storage.
        """
        obj = self.tasks.pop(task_id, None)
        if flush and obj is not None:
            self.flush()

    def _recursive_remove_from_store(self, task_id):
        """
        Remove a task from the store and, if the object has a `tasks`
        attribute containing a list of other tasks, remove them from the store
        """
        queue = [task_id]
        while queue:
            id_to_remove = queue.pop()
            obj = self.store.load(id_to_remove)
            try:
                for child in obj.tasks:
                    queue.append(child.persistent_id)
            except AttributeError:
                pass
            try:
                self.store.remove(id_to_remove)
            except Exception as err:
                gc3libs.log.warning(
                    "Error removing task id `%s` from the store: %s",
                    err)

    def remove(self, task_id, flush=True):
        """
        Remove task identified by `task_id` from the current session
        *and* from the associated storage.
        """
        self._recursive_remove_from_store(task_id)
        self.forget(task_id, flush)

    def __len__(self):
        return len(self.tasks)

    def __iter__(self):
        return iter(self.tasks.values())

    class _BfsIterator(object):
        """
        Iterate over tasks in a group, recursively descending `TaskCollection`s.

        Iteration is done in a breadth-first manner.
        """
        __slots__ = ('_queue')

        def __init__(self, toplevel=None):
            self._queue = []
            if toplevel:
                self._queue.extend(toplevel)

        def __iter__(self):
            return self

        def __next__(self):
            try:
                task = self._queue.pop(0)
                try:
                    self._queue.extend(task.tasks)
                except AttributeError:
                    pass  # `task` is no `TaskCollection`
                return task
            except IndexError: # empty queue
                raise StopIteration()

        next = __next__

    def iter_workflow(self):
        """
        Iterate over all tasks in this session.

        Recursively descends `TaskCollection`:class:
        in  a breadth-first manner.
        """
        return self._BfsIterator(self.tasks.values())

    def list_ids(self):
        """
        Return set of all task IDs belonging to this session.
        """
        return list(self.tasks.keys())

    def list_names(self):
        """
        Return set of names of tasks belonging to this session.
        """
        return set(task.jobname for task in list(self.tasks.values()))

    # persistence management

    def flush(self):
        """
        Update session metadata.

        Should be used after a save/remove operations, to ensure that
        the session state and metadata is correctly persisted.
        """
        # create directory if it does not exists
        if not os.path.exists(self.path):
            os.mkdir(self.path)
        # Update store.url and job_ids.db files
        self._save_store_url_file()
        self._save_index_file()

    def load(self, obj_id, add=True, flush=True):
        """
        Load an object from persistent storage and return it.

        This is just a convenience proxy for calling method `load` on
        the `Store` instance associated with this session.
        """
        obj = self.store.load(obj_id)
        if add:
            self.add(obj, flush)
        return obj

    def load_many(self, obj_ids, add=True, flush=True):
        """
        Load objects given their IDs from persistent storage.

        Return a dictionary mapping task ID to the actual
        retrieved `Task`:class: object.
        """
        tasks = {}
        for task_id in obj_ids:
            try:
                tasks[task_id] = self.load(task_id, add, flush)
            except Exception as err:
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'session',
                        # - class
                        'Session',
                        # - method
                        'load',
                        # - actual error class
                        err.__class__.__name__,
                        # - additional keywords
                        'persistence',
                ):
                    gc3libs.log.warning(
                        "Ignoring error from loading '%s': %s", task_id, err)
                else:
                    # propagate exception back to caller
                    raise
        return tasks

    def save(self, obj):
        """
        Save an object to the persistent storage and return
        `persistent_id` of the saved object.

        This is just a convenience proxy for calling method `save` on
        the `Store` instance associated with this session.

        The object is *not* added to the session, nor is session
        meta-data updated::

            # create an empty session
            >>> import tempfile; tmpdir = tempfile.mktemp(dir='.')
            >>> session = Session(tmpdir)
            >>> 0 == len(session)
            True

            # use `save` on an object
            >>> obj = gc3libs.persistence.Persistable()
            >>> oid = session.save(obj)

            # session is still empty
            >>> 0 == len(session)
            True

            # do cleanup
            >>> session.destroy()
            >>> os.path.exists(session.path)
            False
        """
        return self.store.save(obj)

    def save_all(self, flush=True):
        """
        Save all modified tasks to persistent storage.
        """
        for task in self.tasks.values():
            if task.changed:
                self.save(task)
        if flush:
            self.flush()

    def _save_index_file(self):
        """
        Save job IDs to the default session index.
        """
        idx_filename = os.path.join(self.path, self.INDEX_FILENAME)
        with open(idx_filename, 'w') as idx_fd:
            for task_id in self.tasks:
                idx_fd.write(str(task_id))
                idx_fd.write('\n')

    def _save_store_url_file(self):
        """
        Save the storage URL to a session file.

        If the destination file does not exists, it will
        be created; if it exists, it will be overwritten.
        """
        store_url_filename = os.path.join(self.path, self.STORE_URL_FILENAME)
        gc3libs.utils.write_contents(store_url_filename, str(self.store_url))

    def _touch_file(self, filename, time=None):
        """
        Touch a file which is inside the session directory, updating
        modification and access time to `time` and creating the file
        if it does not exists.

        If `time` is None, the current time will be used.

        It returns the value used for `time`
        """
        filename = os.path.join(self.path, filename)
        open(filename, 'a').close()
        if time is None:
            os.utime(filename, None)
        else:
            os.utime(filename, (time, time))
        return os.stat(filename).st_mtime

    def set_start_timestamp(self, time=None):
        """
        Create a file named `created` in the session directory. It's
        creation/modification time will be used to know when the
        session has sarted.
        """
        self.created = self._touch_file(self.TIMESTAMP_FILES['start'], time)

    def set_end_timestamp(self, time=None):
        """
        Create a file named `finished` in the session directory. It's
        creation/modification time will be used to know when the
        session has finished.

        Please note that `Session` does not know when a session is
        finished, so this method should be called by a
        `SessionBasedScript`:class: class.
        """
        self.finished = self._touch_file(self.TIMESTAMP_FILES['end'], time)


class TemporarySession(Session):
    """
    Create a session from a store URL.

    In contrast with the regular `Session`:class: object, a
    `TemporarySession`:class: does not persist any metadata about the
    task collection.
    In particular:

    - The session index (list of task IDs belonging to the session) is
      initialized from the entire list of jobs present in the given
      `Store`:class: (unless a list is explicitly passed in the
      `task_ids` argument to the constructor). This means that, unlike
      plain `Session`:class: objects, two `TemporarySession`:class:
      objects cannot share the same store.

    - The session directory (``path`` in the `Session`:class:
      constructor) is created on a temporary location on the
      filesystem and deleted when the :class:`TemporarySession` is
      destroyed.

    - Timestamps will be set to the time the `TemporarySession`:class:
      Python object is created; two `TemporarySession`:class:
      instances with the same backing store can have different
      creation timestamps, depending on when exactly they were
      instanciated.

    The `TemporarySession`:class: is only provided as a convenience to
    use code that was built on top of a `Session`:class: with a
    "naked" `Store`:class:.
    """

    def __init__(self, store_or_url, task_ids=None, delete=True, **extra_args):
        # make temporary session dir
        path = tempfile.mkdtemp(
            prefix='gc3pie.TemporarySession.',
            suffix='.d')
        # ensure temp directory is deleted
        if delete:
            def cleanup():
                if os.path.exists(path):
                    shutil.rmtree(path)
            atexit.register(cleanup)
        # init `Session` class
        extra_args.setdefault('create', True)
        extra_args.setdefault('load', False)
        super(TemporarySession, self).__init__(
            path, store_or_url=store_or_url, **extra_args)
        # populate index
        if task_ids is None:
            try:
                task_ids = self.store.list()
            except NotImplementedError:
                raise RuntimeError(
                    "Cannot create temporary session:"
                    " Task store `{0}` does not support listing all task IDs.")
        self.tasks = self.load_many(task_ids, flush=True)
        self._save_index_file()
        # use url as the session name
        self.name = str(self.store.url)


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="filesystem",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python

"""
Run applications as processes starting them from the shell.
"""

# Copyright (C) 2009-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# make coding more python3-ish, must be the first statement
from __future__ import (absolute_import, division, print_function)

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object
from future.utils import with_metaclass

# module doc and other metadata
__docformat__ = 'reStructuredText'


# imports and other dependencies

# stdlib imports
from abc import ABCMeta, abstractmethod
from collections import defaultdict
import pickle as pickle
from getpass import getuser
import os
import os.path
import posixpath
import time

from pkg_resources import Requirement

# GC3Pie imports
import gc3libs
import gc3libs.exceptions
import gc3libs.backends.transport
from gc3libs import log, Run
import gc3libs.defaults
from gc3libs.utils import same_docstring_as, Struct, sh_quote_safe, sh_quote_unsafe
from gc3libs.backends import LRMS
from gc3libs.quantity import Duration, Memory, MB


## helper functions
#
# Mainly for parsing output of shell programs.
#

def _parse_process_status(pstat):
    """
    Map `ps` process status letter to a `Run.State` label.

    Running:

        R: in run queue
        S: interruptible sleep
        D: uninterruptible sleep (Linux)
        U: uninterruptible sleep (MacOSX)
        I: idle (= sleeping > 20s, MacOSX)
        W: paging (Linux, no longer valid since the 2.6.xx kernel)
        Z: "zombie" process

    Stopped:

        T: stopped by job control signal
        t: stopped by debugger during the tracing
        X: dead (should never be seen)

    """
    # Check manpage of ``ps`` both on linux and MacOSX/BSD to know the meaning
    # of these statuses
    if pstat.startswith(('R', 'S', 'D', 'Y', 'I', 'W', 'Z')):
        return Run.State.RUNNING
    elif pstat.startswith(('T', 't', 'X')):
        return Run.State.STOPPED
    else:
        raise KeyError("Unknown process status code `{0}`".format(pstat[0]))


def _parse_time_duration(val):
    """
    Convert the output of common Linux/UNIX system utilities into a GC3Pie `Duration` object.

    Any of the time formats *DD-HH:MM:SS* (days, hours, minutes, seconds),
    *HH:MM:SS* (hours, minutes, seconds), or *MM:SS* (minutes, seconds), or
    even just the number of seconds are acceptable::

      >>> _parse_time_duration('25-00:31:05') == Duration('25d') + Duration('31m') + Duration('5s')
      True

      >>> _parse_time_duration('1:02:03') == Duration('1h') + Duration('2m') + Duration('3s')
      True

      >>> _parse_time_duration('01:02') == Duration('1m') + Duration('2s')
      True

      >>> _parse_time_duration('42') == Duration(42, unit=Duration.s)
      True

    The *seconds* portion of the time string can be followed by
    decimal digits for greater precision::

      >>> _parse_time_duration('0:00.00') == Duration(0, unit=Duration.s)
      True

      >>> _parse_time_duration('4.20') == Duration(4.20, unit=Duration.s)
      True

    When only the number of seconds is given, an optional trailing
    unit specified `s` is allowed::

      >>> _parse_time_duration('4.20s') == Duration(4.20, unit=Duration.s)
      True

    Among the programs whose output can be parsed by this function, there are:

    - GNU time's `%e` format specifier;
    - output of `ps -o etime=` (on both GNU/Linux and MacOSX)
    """
    n = val.count(':')
    if 2 == n:
        if '-' in val:
            days, timespan = val.split('-')
            return (Duration(days + 'd') + Duration(timespan))
        else:
            # Duration's ctor can natively parse this
            return Duration(val)
    elif 1 == n:
        # AA:BB is rejected as ambiguous by `Duration`'s built-in
        # parser; work around it
        mm, ss = val.split(':')
        return (Duration(int(mm, 10), unit=Duration.m)
                + Duration(float(ss), unit=Duration.s))
    elif 0 == n:
        # remove final unit spec, if present
        if val.endswith('s'):
            val = val[:-1]
        # number of seconds with up to 2 decimal precision
        return Duration(float(val), unit=Duration.s)
    else:
        raise ValueError(
            "Expecting duration in the form HH:MM:SS, MM:SS,"
            " or just number of seconds,"
            " got {val} instead".format(val=val))


def _parse_percentage(val):
    """
    Convert a percentage string into a Python float.
    The percent sign at the end is optional::

    >>> _parse_percentage('10')
    10.0
    >>> _parse_percentage('10%')
    10.0
    >>> _parse_percentage('0.25%')
    0.25
    """
    return float(val[:-1]) if val.endswith('%') else float(val)


def _parse_returncode_string(val):
    return Run.shellexit_to_returncode(int(val))


## interface to different OS
#
#

class _Machine(with_metaclass(ABCMeta, object)):
    """
    Base class for OS-specific shell services.

    Linux and MacOSX require slightly different command incantations
    to achieve the same task. The `Machine` class abstract these into
    a uniform interface.
    """

    def __init__(self, transport):
        self.transport = transport

    @staticmethod
    def detect(transport):
        """Factory method to create a `_Machine` instance based on the running kernel."""
        exit_code, stdout, stderr = transport.execute_command('uname -s')
        running_kernel = stdout.strip()
        if running_kernel == 'Linux':
            return _LinuxMachine(transport)
        elif running_kernel == 'Darwin':
            return _MacOSXMachine(transport)
        else:
            raise RuntimeError(
                "Unexpected kernel name: got {0},"
                " expecting one of 'Linux', 'Darwin'"
                .format(running_kernel))

    def _run_command(self, cmd):
        """
        Run a command and return its STDOUT, or raise exception if it errors out.

        This is like `subprocess.check_call`, with the following differences:

        * the slave command is executed through `Transport.execute_command`
          so it need not be locally executed.
        * a `RuntimeError` exception is raised if either the exit code from the
          process is non-zero or STDERR contains any text.
        """
        exit_code, stdout, stderr = self.transport.execute_command(cmd)
        if exit_code != 0 or stderr:
            raise RuntimeError("Got error running command `{0}` (exit code {1}): {2}"
                               .format(cmd, exit_code, stderr.strip()))
        return stdout

    def get_architecture(self):
        cmd = 'uname -m'
        stdout = self._run_command(cmd)
        return gc3libs.config._parse_architecture(stdout)

    def get_process_state(self, pid):
        """
        Return the 1-letter state of process PID.
        (See the ``ps`` man page for a list of possible codes and explanations.)

        Raise ``LookupError`` if no process is identified by the given PID.
        """
        cmd = 'ps -p {0} -o state='.format(pid)
        rc, stdout, stderr = self.transport.execute_command(cmd)
        if rc == 1:
            raise LookupError('No process with PID {0}'.format(pid))
        elif rc == 0:
            return stdout.strip()
        else:
            raise RuntimeError("Got error running command `{0}` (exit code {1}): {2}"
                               .format(cmd, exit_code, stderr.strip()))

    def get_process_running_time(self, pid):
        """
        Return elapsed time since start of process identified by PID.

        Raise ``LookupError`` if no process is identified by the given PID.
        """
        cmd = 'ps -p {0} -o etime='.format(pid)
        rc, stdout, stderr = self.transport.execute_command(cmd)
        if rc == 1:
            raise LookupError('No process with PID {0}'.format(pid))
        elif rc == 0:
            etime = stdout.strip()
            return _parse_time_duration(etime)
        else:
            raise RuntimeError("Got error running command `{0}` (exit code {1}): {2}"
                               .format(cmd, exit_code, stderr.strip()))

    def get_total_cores(self):
        """Return total nr. of CPU cores."""
        cmd = self._get_total_cores_command()
        stdout = self._run_command(cmd)
        try:
            return int(stdout)
        except (ValueError, TypeError) as err:
            raise RuntimeError("Cannot parse output `{0}` of command `{1}`"
                               " as total number of CPU cores: {2}"
                               .format(stdout.strip(), cmd, err))

    # This could be a property of the class, but then we won't be able to
    # use the `@abstractmethod` constructor to enforce that derived classes
    # provide an implementation
    @abstractmethod
    def _get_total_cores_command(self):
        """Command to run to print the nr. of CPU cores to STDOUT."""
        pass

    def get_total_memory(self):
        """
        Return amount of total amount of RAM as a `gc3libs.quantity.Memory` object.
        """
        parts, index, unit = self._get_total_memory_impl()
        try:
            qty = parts[index]
            amount = int(qty)
            return amount * unit
        except KeyError:  # index out of bounds
            raise AssertionError(
                "Call to {0} returned out-of-bounds index {1} into sequence {2}"
                .format(self._get_total_memory_impl, index, parts))
        except (ValueError, TypeError) as err:
            raise RuntimeError("Cannot `{0}` as a memory amount: {1}"
                               .format(qty, err))

    @abstractmethod
    def _get_total_memory_impl(self):
        """Machine-specific part of `get_total_memory`."""
        pass

    def list_process_tree(self, root_pid="1"):
        """
        Return list of PIDs of children of the given process.

        The returned list is empty if no process whose PID is
        `root_pid` can be found.

        Otherwise, the list is composed by walking the tree
        breadth-first, so it always starts with `root_pid` and ends
        with leaf processes (i.e., those which have no children).
        """
        ps_output = self._run_command(self._list_pids_and_ppids_command())

        children = defaultdict(list)
        for line in ps_output.split('\n'):
            line = line.strip()
            if not line:
                continue
            pid, ppid = line.split()
            children[str(ppid)].append(str(pid))
        if root_pid not in children:
            return []

        result = []
        queue = [root_pid]
        while queue:
            node = queue.pop()  # dequeue
            result.append(node)
            for child in children[node]:
                queue.insert(0, child)  # enqueue

        return result


class _LinuxMachine(_Machine):
    """Linux-specific shell tools."""

    def _get_total_cores_command(self):
        """Return nr. of CPU cores from ``nproc``"""
        return 'nproc'

    def _get_total_memory_impl(self):
        with self.transport.open('/proc/meminfo', 'r') as fd:
            for line in fd:
                if line.startswith('MemTotal'):
                    return (line.split(), 1,  Memory.KiB)

    def _list_pids_and_ppids_command(self):
        return 'ps --no-header -o pid,ppid'


class _MacOSXMachine(_Machine):
    """MacOSX-specific shell tools."""

    def _get_total_cores_command(self):
        """Return nr. of CPU cores from ``sysctl hw.ncpu``"""
        return 'sysctl -n hw.ncpu'

    def _get_total_memory_impl(self):
        """Return amount of total memory from ``sysctl hw.memsize``"""
        cmd = 'sysctl hw.memsize'
        stdout = self._run_command(cmd)
        return (stdout.split(':'), 1, Memory.B)

    def _list_pids_and_ppids_command(self):
        return 'ps -o pid=,ppid='


## the main LRMS class
#
#
#

class ShellcmdLrms(LRMS):
    """
    Execute an `Application`:class: instance through the shell.

    Construction of an instance of `ShellcmdLrms` takes the following
    optional parameters (in addition to any parameters taken by the
    base class `LRMS`:class:):

    :param str time_cmd:
      Path to the GNU ``time`` command.  Default is
      `/usr/bin/time`:file: which is correct on all known Linux
      distributions.

      This backend uses many of the
      extended features of GNU ``time``, so the shell-builtins or the
      BSD ``time`` will not work.

    :param str spooldir:
      Path to a filesystem location where to create
      temporary working directories for processes executed through
      this backend. The default value `None` means to use ``$TMPDIR``
      or `/var/tmp`:file: (see `tempfile.mkftemp` for details).

    :param str resourcedir:
      Path to a filesystem location where to create a temporary
      directory that will contain information on the jobs running on
      the machine. The default value `None` means to use
      ``$HOME/.gc3/shellcmd.d``.

    :param str transport:
      Transport to use to connect to the resource. Valid values are
      ``'ssh'`` or ``'local'``.

    :param str frontend:
      If `transport` is ``'ssh'``, then `frontend` is the hostname of the
      remote machine where the jobs will be executed.

    :param bool ignore_ssh_host_key:
      When connecting to a remote resource using the ``'ssh'`` transport the
      server's SSH public key is usually checked against a database of known
      hosts, and if the key is found but it does not match with the one saved
      in the database, the connection will fail. Setting `ignore_ssh_host_key`
      to `True` will disable this check, thus introducing a potential security
      issue but allowing connection even though the database contains
      old/invalid keys. (The main use case is when connecting to VMs on a IaaS
      cloud, since the IP is usually reused and therefore the ssh key is
      recreated.)

    :param bool override:
      `ShellcmdLrms` by default will try to gather information on the
      machine the resource is running on, including the number of
      cores and the available memory. These values may be different
      from the values stored in the configuration file. If `override`
      is ``True``, then the values automatically discovered will be used
      instead of the ones in the configuration file. If `override` is
      False, instead, the values in the configuration file will be
      used.

    :param int ssh_timeout:
      If `transport` is ``'ssh'``, this value will be used as timeout (in
      seconds) for connecting to the SSH TCP socket.

    :param gc3libs.quantity.Memory large_file_threshold:
      Copy files below this size in one single SFTP GET operation;
      see `SshTransport.get`:meth: for more information.
      Only used if `transport` is ``'ssh'``.

    :param gc3libs.quantity.Memory large_file_chunk_size:
      Copy files that are over the above-mentioned threshold by
      sequentially transferring chunks of this size.
      see `SshTransport.get`:meth: for more information.
      Only used if `transport` is ``'ssh'``.
    """

    TIMEFMT = '\n'.join([
        'WallTime=%es',
        'KernelTime=%Ss',
        'UserTime=%Us',
        'CPUUsage=%P',
        'MaxResidentMemory=%MkB',
        'AverageResidentMemory=%tkB',
        'AverageTotalMemory=%KkB',
        'AverageUnsharedMemory=%DkB',
        'AverageUnsharedStack=%pkB',
        'AverageSharedMemory=%XkB',
        'PageSize=%ZB',
        'MajorPageFaults=%F',
        'MinorPageFaults=%R',
        'Swaps=%W',
        'ForcedSwitches=%c',
        'WaitSwitches=%w',
        'Inputs=%I',
        'Outputs=%O',
        'SocketReceived=%r',
        'SocketSent=%s',
        'Signals=%k',
        'ReturnCode=%x',
    ])
    """
    Format string for running commands with ``/usr/bin/time``.
    It is used by GC3Pie to capture resource usage data for commands
    executed through the shell.

    The value used here lists all the resource usage values that *GNU
    time* can capture, with the same names used by the ARC Resource
    Manager (for historical reasons).
    """

    # how to translate GNU time output into GC3Pie execution metrics
    TIMEFMT_CONV = {
        # GNU time output key    .execution attr                     converter function
        # |                        |                                  |
        # v                        v                                  v
        'WallTime':              ('duration',                         _parse_time_duration),
        'KernelTime':            ('shellcmd_kernel_time',             Duration),
        'UserTime':              ('shellcmd_user_time',               Duration),
        'CPUUsage':              ('shellcmd_cpu_usage',               _parse_percentage),
        'MaxResidentMemory':     ('max_used_memory',                  Memory),
        'AverageResidentMemory': ('shellcmd_average_resident_memory', Memory),
        'AverageTotalMemory':    ('shellcmd_average_total_memory',    Memory),
        'AverageUnsharedMemory': ('shellcmd_average_unshared_memory', Memory),
        'AverageUnsharedStack':  ('shellcmd_average_unshared_stack',  Memory),
        'AverageSharedMemory':   ('shellcmd_average_shared_memory',   Memory),
        'PageSize':              ('shellcmd_page_size',               Memory),
        'MajorPageFaults':       ('shellcmd_major_page_faults',       int),
        'MinorPageFaults':       ('shellcmd_minor_page_faults',       int),
        'Swaps':                 ('shellcmd_swapped',                 int),
        'ForcedSwitches':        ('shellcmd_involuntary_context_switches', int),
        'WaitSwitches':          ('shellcmd_voluntary_context_switches',   int),
        'Inputs':                ('shellcmd_filesystem_inputs',       int),
        'Outputs':               ('shellcmd_filesystem_outputs',      int),
        'SocketReceived':        ('shellcmd_socket_received',         int),
        'SocketSent':            ('shellcmd_socket_sent',             int),
        'Signals':               ('shellcmd_signals_delivered',       int),
        'ReturnCode':            ('returncode',                       _parse_returncode_string),
    }
    """
    How to translate *GNU time* output into values stored in the ``.execution`` attribute.

    The dictionary maps key names (as used in the `TIMEFMT` string) to
    a pair *(attribute name, converter function)* consisting of the
    name of an attribute that will be set on a task's ``.execution``
    object, and a function to convert the (string) value gotten from
    *GNU time* output into the actual Python value written.
    """

    PRIVATE_DIR = '.gc3pie_shellcmd'
    """
    Subdirectory of a tasks's execution directory reserved for storing
    `ShellcmdLrms`:class: files.
    """

    WRAPPER_SCRIPT = 'wrapper_script.sh'
    """
    Name of the task launcher script (within `PRIVATE_DIR`).

    The `ShellcmdLrms`:class: writes here that wrap an application's
    payload script, to collect resource usage or download/upload
    result files, etc.
    """

    WRAPPER_OUTPUT_FILENAME = 'resource_usage.txt'
    """
    Name of the file where resource usage is written to.

    (Relative to `PRIVATE_DIR`.)
    """

    WRAPPER_PID = 'wrapper.pid'
    """
    Name of the file where the wrapper script's PID is stored.

    (Relative to `PRIVATE_DIR`).
    """

    MOVER_SCRIPT = 'mover.py'
    """
    Name of the data uploader/downloader script (within `PRIVATE_DIR`).
    """

    RESOURCE_DIR = '$HOME/.gc3/shellcmd.d'
    """
    Path to the directory where bookkeeping files are stored.
    (This is on the target machine where `ShellcmdLrms`:class:
    executes commands.)

    It may contain environmental variable references, which are
    expanded through the (remote) shell.
    """

    def __init__(self, name,
                 # these parameters are inherited from the `LRMS` class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth=None,
                 # these are specific to `ShellcmdLrms`
                 frontend='localhost', transport='local',
                 time_cmd=None,
                 override='False',
                 spooldir=gc3libs.defaults.SPOOLDIR,
                 resourcedir=None,
                 # SSH-related options; ignored if `transport` is 'local'
                 ssh_config=None,
                 keyfile=None,
                 ignore_ssh_host_keys=False,
                 ssh_timeout=None,
                 large_file_threshold=None,
                 large_file_chunk_size=None,
                 **extra_args):

        # init base class
        LRMS.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth, **extra_args)

        # whether actual machine params (cores, memory) should be
        # auto-detected on first use
        self.override = override

        self.spooldir = spooldir

        # Configure transport
        if transport == 'local':
            self.transport = gc3libs.backends.transport.LocalTransport()
            self._username = getuser()
            self.frontend = 'localhost'
        elif transport == 'ssh':
            auth = self._auth_fn()
            self._username = auth.username
            self.transport = gc3libs.backends.transport.SshTransport(
                frontend,
                ignore_ssh_host_keys=ignore_ssh_host_keys,
                ssh_config=(ssh_config or auth.ssh_config),
                username=self._username,
                port=auth.port,
                keyfile=(keyfile or auth.keyfile),
                timeout=(ssh_timeout or auth.timeout),
                large_file_threshold=large_file_threshold,
                large_file_chunk_size=large_file_chunk_size,
            )
            self.frontend = frontend
        else:
            raise AssertionError("Unknown transport '{0}'" .format(transport))

        # Init bookkeeping
        self.updated = False  # data may not reflect actual state
        self.free_slots = self.max_cores
        self.user_run = 0
        self.user_queued = 0
        self.queued = 0
        self.total_memory = max_memory_per_core
        self.available_memory = self.total_memory
        self._job_infos = {}

        # Some init parameters can only be discovered / checked when a
        # connection to the target resource is up.  We want to delay
        # this until the time they are actually used, to avoid opening
        # up a network connection when the backend is initialized
        # (could be e.g. a `ginfo -n` call that has no need to operate
        # on remote objects).
        self._resourcedir_raw = resourcedir or ShellcmdLrms.RESOURCE_DIR
        self._time_cmd = time_cmd
        self._time_cmd_ok = False  # check on first use

    @property
    def frontend(self):
        return self._frontend

    @frontend.setter
    def frontend(self, value):
        self._frontend = value
        self.transport.set_connection_params(value)

    @property
    def resource_dir(self):
        try:
            return self._resource_dir
        except AttributeError:
            # Since RESOURCE_DIR contains the `$HOME` variable, we
            # have to expand it by connecting to the remote
            # host. However, we don't want to do that during
            # initialization, so we do it the first time this is
            # actually needed.
            self._init_resource_dir()
            return self._resource_dir

    @resource_dir.setter
    def resource_dir(self, value):
        self._resource_dir = value

    def _init_resource_dir(self):
        self.transport.connect()
        # expand env variables in the `resource_dir` setting
        exit_code, stdout, stderr = self.transport.execute_command(
            'echo %s' % sh_quote_unsafe(self._resourcedir_raw))
        self.resource_dir = stdout.strip()

        if not self.transport.exists(self.resource_dir):
            try:
                log.info("Creating resource directory: '%s' ...", self.resource_dir)
                self.transport.makedirs(self.resource_dir)
            except Exception as ex:
                log.error("Failed creating resource directory '%s': %s: %s",
                          self.resource_dir, type(ex), ex)
                # cannot continue
                raise

    @property
    def time_cmd(self):
        if not self._time_cmd_ok:
            self._time_cmd = self._locate_gnu_time()
            self._time_cmd_ok = True
        return self._time_cmd

    def _gather_machine_specs(self):
        """
        Gather information about target machine and update config.
        The following attributes are set (or reset) as an effect
        of calling this method:

        - ``_machine``: Set to the an appropriate instance of
          `_Machine`:class:, detected through connection via
          `self.transport`.
        - ``_resource_dir``: Set to the expansion of whatever was
          passed as ``resource`` construction parameter.
        - ``max_cores``: If ``self.override`` is true, set to the
          number of processors on the target.
        - ``total_memory``: If ``self.override`` is true, set to total
          amount of memory on the target.
        """
        self.transport.connect()
        self._machine = _Machine.detect(self.transport)
        self._init_arch()
        if self.override:
            self._init_max_cores()
            self._init_total_memory()
            self._update_resource_usage_info()

    def _init_arch(self):
        arch = self._machine.get_architecture()
        if not (arch <= self.architecture):
            if self.override:
                log.info(
                    "Mismatch of value `architecture` on resource %s:"
                    " configuration file says `architecture=%s`"
                    " but GC3Pie detected `%s`. Updating current value.",
                    self.name,
                    ','.join(self.architecture),
                    ','.join(arch))
                self.architecture = arch
            else:
                raise gc3libs.exceptions.ConfigurationError(
                    "Invalid architecture: configuration file says `%s` but "
                    "it actually is `%s`" % (', '.join(self.architecture),
                                             ', '.join(arch)))

    def _init_max_cores(self):
        max_cores = self._machine.get_total_cores()
        if max_cores != self.max_cores:
            log.info(
                "Mismatch of value `max_cores` on resource '%s':"
                " configuration file says `max_cores=%d` while it's actually `%d`."
                " Updating current value.",
                self.name, self.max_cores, max_cores)
            self.max_cores = max_cores

    def _init_total_memory(self):
        self.total_memory = self._machine.get_total_memory()
        if self.total_memory != self.max_memory_per_core:
            log.info(
                "Mismatch of value `max_memory_per_core` on resource %s:"
                " configuration file says `max_memory_per_core=%s` while it's"
                " actually `%s`. Updating current value.",
                self.name,
                self.max_memory_per_core,
                self.total_memory.to_str('%g%s', unit=Memory.MB))
            self.max_memory_per_core = self.total_memory

    def _locate_gnu_time(self):
        """
        Return the command path to run the GNU `time` binary.

        :raise ConfigurationError:
          if no GNU ``time`` executable can be located.
        """
        candidates = [
            'time',  # default on Linux systems
            'gtime', # MacOSX with Homebrew or MacPorts
        ]
        if self._time_cmd:
            # try this first
            candidates.insert(0, self._time_cmd)
        for time_cmd in candidates:
            gc3libs.log.debug(
                "Checking if GNU time is available as command `%s`", time_cmd)
            # We use `command` in order to force the shell to execute
            # the binary and not the shell builtin (cf. the POSIX
            # standard).  However note that the wrapper script will
            # execute `exec time_cmd` in order to replace the current
            # shell, but `exec` will never run the builtin.
            exit_code, stdout, stderr = self.transport.execute_command(
                'command %s --version 2>&1 | grep GNU' % time_cmd)
            if exit_code == 0:
                # command is GNU! Good!
                return time_cmd
        raise gc3libs.exceptions.ConfigurationError(
            "Unable to find GNU `time` on resource `{name}`."
            " Please, install GNU time and set the `time_cmd`"
            " configuration option in gc3pie.conf."
            .format(name=self.name))

    ## Bookkeeping
    #
    # The following methods deal with internal book-keeping: how much
    # of the target's configured resources have been used by GC3Pie.
    # Presently, book-keeping is so complicated (and requires
    # recomputing at each invocation) because GC3Pie makes the
    # assumption that the target resource is *shared*, i.e., other
    # GC3Pie processes run concurrently by the user may compete for
    # the same resources.
    #

    def count_running_tasks(self):
        """
        Returns number of currently running tasks.

        .. note::

          1. The count of running tasks includes also tasks that may
             have been started by another GC3Pie process so this count
             can be positive when the resource has just been opened.

          2. The count is updated every time the resource is updated,
             so the returned number can be stale if the
             `ShellcmdLrms.get_resource_status()` has not been called
             for a while.
        """
        return sum(1 for info in list(self._job_infos.values())
                   if not info['terminated'])

    def count_used_cores(self):
        """
        Return total nr. of cores used by running tasks.

        Similar caveats as in `ShellcmdLrms.count_running_tasks`:meth:
        apply here.
        """
        return sum(info['requested_cores']
                   for info in list(self._job_infos.values())
                   if not info['terminated'])

    def count_used_memory(self):
        """
        Return total amount of memory used by running tasks.

        Similar caveats as in `ShellcmdLrms.count_running_tasks`:meth:
        apply here.
        """
        return sum(
            # FIXME: if `requested_memory==None` then just do not
            # account a task's memory usage.  This is of course
            # incorrect and leads to situations where a single task
            # can wreak havoc on an entire compute node, but is
            # consistent with what we do during scheduling /
            # requirements check.  (OTOH, it's *not* consistent with
            # what other backends do: SLURM, for example, takes the
            # pessimistic stance that a job with no memory
            # requirements is using (DefMemPerCPU * NumCPUs)
            ((info['requested_memory'] or 0*MB)
             for info in list(self._job_infos.values())
             if not info['terminated']), 0*MB)


    def _get_persisted_job_info(self):
        """
        Get information on total resources from the files stored in
        `self.resource_dir`. It then returns a dictionary {PID: {key:
        values}} with informations for each job which is associated to
        a running process.
        """
        self.transport.connect()
        job_infos = {}
        pidfiles = self.transport.listdir(self.resource_dir)
        if pidfiles:
            log.debug("Checking status of the following PIDs: %s",
                      ", ".join(pidfiles))
            for pid in pidfiles:
                job = self._read_job_info_file(pid)
                if job:
                    job_infos[str(pid)] = job
                else:
                    # Process not found, ignore it
                    continue
        return job_infos

    def _read_job_info_file(self, pid):
        """
        Get resource information on job with pid `pid`, if it
        exists. Returns None if it does not exist.
        """
        self.transport.connect()
        log.debug("Reading job info file for pid %r", pid)
        jobinfo = None
        path = posixpath.join(self.resource_dir, str(pid))
        with self.transport.open(path, 'rb') as fp:
            try:
                jobinfo = pickle.load(fp)
            except Exception as ex:
                log.error("Unable to read remote job info file %s: %s",
                          path, ex)
                raise
        return jobinfo

    def _write_job_info_file(self, pid, resources):
        """
        Update file in `self.resource_dir/PID` with `resources`.
        """
        self.transport.connect()
        # XXX: We should check for exceptions!
        log.debug("Updating job info file for pid %s", pid)
        with self.transport.open(
                posixpath.join(self.resource_dir, str(pid)), 'wb') as fp:
            pickle.dump(resources, fp, -1)

    def _delete_job_info_file(self, pid):
        """
        Delete `self.resource_dir/PID` file
        """
        self.transport.connect()
        log.debug("Deleting job info file for pid %s ...", pid)
        pidfile = posixpath.join(self.resource_dir, pid)
        try:
            self.transport.remove(pidfile)
        except Exception as err:
            msg = str(err)
            if 'OSError: [Errno 2]' not in msg:
                log.debug(
                    "Ignored error deleting file `%s`: %s: %s",
                    pidfile, err.__class__.__name__, err)


    ## Backend interface implementation
    #
    # These methods provide what is expected of any LRMS class.
    #

    def _connect(self):
        """Ensure transport to remote resource works."""
        try:
            self.transport.connect()
        except gc3libs.exceptions.TransportError as err:
            log.error("Unable to connect to host `%s`: %s", self.frontend, err)
            raise
        try:
            self._machine
        except AttributeError:
            self._gather_machine_specs()

    def cancel_job(self, app):
        """
        Kill all children processes of the given task `app`.

        The PID of the wrapper script (which is the root of the PID
        tree we are going to send a "TERM" signal) must have been
        stored (by `submit_job`:meth:) as `app.execution.lrms_jobid`.
        """
        try:
            root_pid = app.execution.lrms_jobid
        except ValueError:
            raise gc3libs.exceptions.InvalidArgument(
                "Invalid field `lrms_jobid` in Task '{0}':"
                " expected a number, got '{{2}) instead"
                .format(app, app.execution.lrms_jobid,
                        type(app.execution.lrms_jobid)))

        self._connect()

        pids_to_kill = self._machine.list_process_tree(root_pid)
        if not pids_to_kill:
            log.debug(
                "No process identified by PID %s in `ps` output,"
                " assuming task %s is already terminated.", root_pid, app)
        else:
            assert (root_pid == pids_to_kill[0])

            # list `pids_to_kill` starts with the root process and ends
            # with leaf ones; we want to kill them in reverse order (leaves first)
            kill_args = ' '.join(str(procid) for procid in reversed(pids_to_kill))
            log.debug(
                "Cancelling task %s on resource `%s`:"
                " sending SIGTERM to processes with PIDs %s",
                app, self.name, kill_args)
            # ignore exit code and STDERR from `kill`: if any process
            # exists while we're killing them, `kill` will error out
            # but that error should be ignored...
            self.transport.execute_command("kill {0}".format(kill_args))

            # now double-check and send SIGKILLs
            attempt = 1
            waited_on_first = False
            # XXX: should these be configurable?
            wait = 60
            max_attempts = 5
            while pids_to_kill and attempt < max_attempts:
                for target in list(reversed(pids_to_kill)):
                    log.debug("Checking if PID %s is still running ...", target)
                    try:
                        pstat = self._machine.get_process_state(target)
                    except LookupError:
                        log.debug("Process %s can no longer be found in process table.", target)
                        pids_to_kill.remove(target)
                        continue
                    # see comments in `_parse_process_status` for the
                    # meaning of the letters
                    if pstat in ['R', 'S', 'I', 'W', 'T', 't']:
                        if not waited_on_first:
                            # wait some time to allow disk I/O before termination
                            self._grace_time(wait)
                            waited_on_first = True
                        exit_code, stdout, stderr = self.transport.execute_command(
                            'kill -9 {0}'.format(target))
                        if exit_code == 0:
                            pids_to_kill.remove(target)
                        else:
                            log.debug(
                                "Could not send SIGKILL"
                                " to process %s on resource '%s':"
                                " %s (exit code: %d)",
                                target, self.name, stderr.strip(), exit_code)
                    elif pstat in ['D', 'U']:
                        log.error(
                            "Process %s on resource %s is"
                            " in uninterruptible sleep and cannot be killed.",
                            target, self.name)
                    elif pstat in ['X', 'Z']:
                        log.warning(
                            "Process %s on resource %s is already dead"
                            " but process entry has not been cleared."
                            " This might be a bug in GC3Pie or in `%s`.",
                                 target, self.name, self.time_cmd)
                        pids_to_kill.remove(target)
                if not pids_to_kill:
                    break
                self._grace_time(wait)
                attempt += 1

            if pids_to_kill:
                log.error(
                    "Not all processes belonging to task %s could be killed:"
                    " processes %s still alive on resource '%s'.",
                    app, (' '.join(str(p) for p in pids_to_kill)), self.name)

        try:
            self._job_infos[root_pid]['terminated'] = True
        except KeyError:
            # It may happen than `cancel_job()` is called without the
            # resource state having been updated (hence
            # `self._job_infos` is empty); this happens e.g. with the
            # `gkill` command.  If that happens, just ignore the
            # error.  (XXX: There might be a better way to handle this...)
            if self.updated:
                raise
            else:
                # ignore
                pass

    def _grace_time(self, wait):
        if wait:
            log.info("Waiting %s seconds to allow processes to terminate ...", wait)
            for _ in range(wait):  # Python ignores SIGINT while in `time.sleep()`
                time.sleep(1)


    @same_docstring_as(LRMS.close)
    def close(self):
        # XXX: free any resources in use?
        pass


    def free(self, app):
        """
        Delete the temporary directory where a child process has run.
        The temporary directory is removed with all its content,
        recursively.

        If deletion is successful, the `lrms_execdir` attribute in
        `app.execution` is reset to `None`; subsequent invocations of
        this method on the same applications do nothing.
        """
        try:
            if (hasattr(app.execution, 'lrms_execdir') and app.execution.lrms_execdir is not None):
                log.debug('Deleting working directory of task `%s` ...', app)
                self.transport.connect()
                if self.transport.isdir(app.execution.lrms_execdir):
                    self.transport.remove_tree(app.execution.lrms_execdir)
                app.execution.lrms_execdir = None
        except Exception as ex:
            log.warning("Could not remove directory '%s': %s: %s",
                        app.execution.lrms_execdir, ex.__class__.__name__, ex)

        try:
            pid = app.execution.lrms_jobid
            self._delete_job_info_file(pid)
        except AttributeError:
            # lrms_jobid not yet assigned; probably submit
            # failed -- ignore and continue
            pass


    @same_docstring_as(LRMS.get_resource_status)
    def get_resource_status(self):
        self.updated = False
        self._connect()
        self._update_resource_usage_info()
        self.updated = True
        return self

    def _update_resource_usage_info(self):
        """
        Helper method for (re)reading resource usage from disk.
        """
        self._job_infos = self._get_persisted_job_info()
        used_memory = self.count_used_memory()
        self.available_memory = self.total_memory - used_memory
        self.free_slots = self.max_cores - self.count_used_cores()
        self.user_run = self.count_running_tasks()
        log.debug("Recovered resource information from files in %s:"
                  " total nr. of cores: %s, requested by jobs: %s;"
                  " available memory: %s, requested by jobs: %s.",
                  self.resource_dir,
                  self.max_cores, (self.max_cores - self.free_slots),
                  self.available_memory.to_str('%g%s', unit=Memory.MB, conv=float),
                  used_memory.to_str('%g%s', unit=Memory.MB, conv=float))


    @same_docstring_as(LRMS.get_results)
    def get_results(self, app, download_dir,
                    overwrite=False, changed_only=True):
        if app.output_base_url is not None:
            raise gc3libs.exceptions.DataStagingError(
                "Retrieval of output files to non-local destinations"
                " is not supported in the ShellCmd backend.")

        self._connect()
        # Make list of files to copy, in the form of (remote_path,
        # local_path) pairs.  This entails walking the
        # `Application.outputs` list to expand wildcards and
        # directory references.
        stageout = list()
        for remote_relpath, local_url in app.outputs.items():
            if local_url.scheme in ['swift', 'swt', 'swifts', 'swts']:
                continue
            local_relpath = local_url.path
            if remote_relpath == gc3libs.ANY_OUTPUT:
                remote_relpath = ''
                local_relpath = ''
            stageout += self._get_remote_and_local_path_pair(
                app, remote_relpath, download_dir, local_relpath)

        # copy back all files, renaming them to adhere to the
        # ArcLRMS convention
        log.debug("Downloading job output into '%s' ...", download_dir)
        for remote_path, local_path in stageout:
            # ignore missing files (this is what ARC does too)
            self.transport.get(remote_path, local_path,
                               ignore_nonexisting=True,
                               overwrite=overwrite,
                               changed_only=changed_only)
        return

    def _get_remote_and_local_path_pair(self, app, remote_relpath,
                                         local_root_dir, local_relpath):
        """
        Scan remote directory and return list of corresponding remote and local paths.

        The return value is a list of *(remote_path, local_path)* pairs: each
        `remote_path` is an existing file on the remote end of the transport,
        and *local_path* is the corresponding local path, constructed by
        prepending `local_root_dir` to the relative remote path.
        """
        # see https://github.com/fabric/fabric/issues/306 about why it is
        # correct to use `posixpath.join` for remote paths (instead of
        # `os.path.join`)
        remote_path = posixpath.join(app.execution.lrms_execdir, remote_relpath)
        local_path = os.path.join(local_root_dir, local_relpath)
        if self.transport.isdir(remote_path):
            # recurse, accumulating results
            result = list()
            for entry in self.transport.listdir(remote_path):
                result += self._get_remote_and_local_path_pair(
                    app, posixpath.join(remote_relpath, entry), local_path, entry)
            return result
        else:
            return [(remote_path, local_path)]


    def has_running_tasks(self):
        """
        Return ``True`` if tasks are running on the resource.

        See `ShellcmdLrms.count_running_tasks`:meth: for caveats
        about the count of "running jobs" upon which this boolean
        check is based.
        """
        return self.user_run > 0


    @same_docstring_as(LRMS.peek)
    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        # `remote_filename` must be relative to the execution directory
        assert not os.path.isabs(remote_filename)
        self._connect()
        # see https://github.com/fabric/fabric/issues/306 about why it
        # is correct to use `posixpath.join` for remote paths (instead
        # of `os.path.join`)
        remote_filename = posixpath.join(app.execution.lrms_execdir, remote_filename)
        statinfo = self.transport.stat(remote_filename)
        # seeking backwards past the beginning of the file makes the
        # read() fail in SFTP, so be sure that we do not rewind too
        # much...
        if offset < 0 and -offset > statinfo.st_size:
            offset = 0
        gc3libs.log.debug(
            "Reading %s bytes starting at offset %s from remote file '%s' ...",
            size, offset, remote_filename)
        with self.transport.open(remote_filename, 'r') as remotely:
            if offset >= 0:
                # seek from start of file
                remotely.seek(offset, os.SEEK_SET)
            else:
                # seek from end of file
                remotely.seek(offset, os.SEEK_END)
            data = remotely.read(size)
            try:
                # is `local_file` a file-like object?
                local_file.write(data)
            except (TypeError, AttributeError):
                # ... no, then treat it as a file name
                with open(local_file, 'w+b') as locally:
                    locally.write(data)


    def submit_job(self, app):
        """
        Run an `Application` instance as a shell process.

        :see: `LRMS.submit_job`
        """
        #
        # This is a rather lengthy (if straightforward) function, whose actual
        # implementation is brokwn down into several smaller pieces, listed
        # hereafter. The body of `submit_jobs()` should give a high-level
        # overview, and the helper functions carry the implementation details.
        #

        # if any failure happens here, we just raise an `LRMSSubmitError` and
        # be done with it: there is nothing to clean up
        try:
            self._connect()
            self._check_app_requirements(app)
        except gc3libs.exceptions.LRMSSubmitError:
            raise  # no need to convert
        except Exception as err:
            raise gc3libs.exceptions.LRMSSubmitError(
                "Failed submitting task {0} to resource `{1}`: {2}"
                .format(app, self.name, err))

        log.debug("Executing command '%s' through the shell on `%s` ...",
                  " ".join(app.arguments), self.frontend)

        # from this point on, any failure needs clean-up by calling `self.free(app)`
        try:
            app.execution.lrms_execdir = self._setup_app_execution_directory(app)
            self._stage_app_input_files(app)
            self._ensure_app_command_is_executable(app)

            redirection_command = self._setup_redirection(app)
            env_commands = self._setup_environment(app)

            wrapper_dir = self._setup_wrapper_dir(app)
            download_cmds, upload_cmds = self._setup_data_movers(app, wrapper_dir)
            pidfilename = posixpath.join(wrapper_dir, self.WRAPPER_PID)
            wrapper_output_path = posixpath.join(wrapper_dir, self.WRAPPER_OUTPUT_FILENAME)
            wrapper_script_path = posixpath.join(wrapper_dir, self.WRAPPER_SCRIPT)

            # create the wrapper script
            with self.transport.open(wrapper_script_path, 'wt') as wrapper:
                wrapper.write(
                    r"""#!/bin/sh
                    echo $$ >'{pidfilename}'
                    cd {execdir}
                    {redirections}
                    {environment}
                    {download_cmds}
                    '{time_cmd}' -o '{wrapper_out}' -f '{fmt}' {command}
                    rc=$?
                    {upload_cmds}
                    rc2=$?
                    if [ $rc -ne 0 ]; then exit $rc; else exit $rc2; fi
                    """.format(
                        pidfilename=pidfilename,
                        execdir=app.execution.lrms_execdir,
                        time_cmd=self.time_cmd,
                        wrapper_out=wrapper_output_path,
                        fmt=ShellcmdLrms.TIMEFMT,
                        redirections=redirection_command,
                        environment=('\n'.join(env_commands)),
                        download_cmds=('\n'.join(download_cmds)),
                        upload_cmds=('\n'.join(upload_cmds)),
                       command=(' '.join(sh_quote_unsafe(arg) for arg in app.arguments)),
                ))
            self.transport.chmod(wrapper_script_path, 0o755)

            # execute the script in background
            self.transport.execute_command(wrapper_script_path, detach=True)
            pid = self._read_app_process_id(pidfilename)
            app.execution.lrms_jobid = pid

        except gc3libs.exceptions.LRMSSubmitError:
            self.free(app)
            raise  # no need to convert

        except Exception as err:
            self.free(app)
            raise gc3libs.exceptions.LRMSSubmitError(
                "Failed submitting task {0} to resource `{1}`: {2}"
                .format(app, self.name, err))

        # Update application and current resources
        #
        self.free_slots -= app.requested_cores
        self.user_run += 1
        if app.requested_memory:
            self.available_memory -= app.requested_memory
        self._job_infos[pid] = {
            'requested_cores': app.requested_cores,
            'requested_memory': app.requested_memory,
            'execution_dir': app.execution.lrms_execdir,
            'terminated': False,
        }
        self._write_job_info_file(pid, self._job_infos[pid])

        return app

    ## implementation of the `submit_job()` method

    def _check_app_requirements(self, app, update=True):
        """Raise exception if application requirements cannot be satisfied."""
        # update resource status to get latest data on free cores, memory, etc
        if update:
            self.get_resource_status()

        if self.free_slots == 0:  # or free_slots == 0:
            if self.override:
                errmsg = (
                    "Resource {0} already running maximum allowed number of jobs"
                    .format(self.name))
            else:
                errmsg = (
                    "Resource {0} already running maximum allowed number of jobs"
                    " ({1}). Increase 'max_cores' to raise."
                    .format(self.name, self.max_cores))
            raise gc3libs.exceptions.MaximumCapacityReached(errmsg)

        if self.free_slots < app.requested_cores:
            raise gc3libs.exceptions.MaximumCapacityReached(
                "Resource {0} does not have enough available execution slots:"
                " {1} requested total, but only {2} available."
                .format(self.name, app.requested_cores, self.free_slots))

        if (app.requested_memory and self.available_memory < app.requested_memory):
            raise gc3libs.exceptions.MaximumCapacityReached(
                "Resource {0} does not have enough available memory:"
                " {1} requested total, but only {2} available."
                .format(
                    self.name,
                    app.requested_memory.to_str('%g%s', unit=Memory.MB),
                    self.available_memory.to_str('%g%s', unit=Memory.MB))
            )

    def _setup_app_execution_directory(self, app):
        """
        Create a temporary subdirectory and return its (remote) path.

        This is intended to be used as part of the "submit" process;
        in case of any failure, raises a `SpoolDirError`.
        """
        target = posixpath.join(self.spooldir, 'shellcmd_job.XXXXXX')
        cmd =  ("mkdir -p {0} && mktemp -d {1}" .format(self.spooldir, target))
        exit_code, stdout, stderr = self.transport.execute_command(cmd)
        if exit_code != 0 or stderr:
            raise gc3libs.exceptions.SpoolDirError(
                "Cannot create temporary job working directory"
                " `{0}` on host `{1}`; command `{2}` exited"
                " with code {3} and error output: `{4}`."
                .format(target, self.frontend, cmd, exit_code, stderr))
        return stdout.strip()

    def _stage_app_input_files(self, app):
        destdir = app.execution.lrms_execdir
        for local_path, remote_path in list(app.inputs.items()):
            if local_path.scheme != 'file':
                log.debug(
                    "Ignoring input URL `%s` for task %s:"
                    " only 'file:' schema supported by the ShellcmdLrms backend.",
                    local_path, app)
                continue
            remote_path = posixpath.join(destdir, remote_path)
            remote_parent = os.path.dirname(remote_path)
            try:
                if (remote_parent not in ['', '.']
                        and not self.transport.exists(remote_parent)):
                    self.transport.makedirs(remote_parent)
                self.transport.put(local_path.path, remote_path)
                # preserve execute permission on input files
                if os.access(local_path.path, os.X_OK):
                    self.transport.chmod(remote_path, 0o755)
            except Exception as err:
                log.error(
                    "Staging input file `%s` to host `%s` failed: %s",
                    local_path.path, self.frontend, err)
                raise

    def _ensure_app_command_is_executable(self, app):
        """
        Give an `Application`'s command file ``a+x`` permissions.

        Ignore all failures: might be a link to a file we do not own.

        In addition, only do this for "local" executable files: by documented
        convention, any command name *not* starting with `./` will be searched
        in `$PATH` and not in the current working directory (i.e., is not
        something staged by GC3Pie).
        """
        execdir = app.execution.lrms_execdir
        cmd = app.arguments[0]
        if cmd.startswith('./'):
            try:
                self.transport.chmod(posixpath.join(execdir, cmd[2:]), 0o755)
            except:
                log.warning(
                    "Failed setting execution flag on remote file '%s'",
                    posixpath.join(execdir, cmd))

    def _setup_redirection(self, app):
        """
        Return shell redirection operators to be applied when executing `app`.

        Also ensure that directories where STDOUT and STDERR files should be
        captured exist (on the remote side).
        """
        execdir = app.execution.lrms_execdir
        redirections = ['exec']
        if app.stdin is not None:
            redirections.append("<'%s'" % app.stdin)
        if app.stdout is not None:
            redirections.append(">'%s'" % app.stdout)
            stdout_dir = os.path.dirname(app.stdout)
            if stdout_dir:
                self.transport.makedirs(posixpath.join(execdir, stdout_dir))
        if app.join or (app.stderr and app.stderr == app.stdout):
            redirections.append("2>&1")
        else:
            if app.stderr is not None:
                redirections.append("2>'%s'" % app.stderr)
                stderr_dir = os.path.dirname(app.stderr)
                if stderr_dir:
                    self.transport.makedirs(posixpath.join(execdir, stderr_dir))
        if len(redirections) > 1:
            return ' '.join(redirections)
        else:
            return ''  # nothing to do

    def _setup_environment(self, app):
        """Return commands to set up the environment for `app`."""
        env_commands = []
        for k, v in app.environment.items():
            env_commands.append(
                "export {k}={v};"
                .format(k=sh_quote_safe(k), v=sh_quote_unsafe(v)))
        return env_commands

    def _setup_wrapper_dir(self, app):
        """
        Create and return the directory in which GC3Pie aux files for `app` will be stored.
        """
        wrapper_dir = posixpath.join(app.execution.lrms_execdir, self.PRIVATE_DIR)
        if not self.transport.isdir(wrapper_dir):
            try:
                self.transport.makedirs(wrapper_dir)
            except Exception as err:
                log.error("Failed creating remote folder '%s': %s", wrapper_dir, err)
                raise
        return wrapper_dir

    # FIXME: data mover functionality should be available in all backends!
    def _setup_data_movers(self, app, destdir):
        """
        Set up scripts to download/upload the swift/http files.
        """
        download_cmds = []
        upload_cmds = []
        mover_path = posixpath.join(destdir, ShellcmdLrms.MOVER_SCRIPT)
        for url, outfile in list(app.inputs.items()):
            if url.scheme in ['swift', 'swifts', 'swt', 'swts', 'http', 'https']:
                download_cmds.append(
                    "python '{mover}' download '{url}' '{outfile}'"
                    .format(mover=mover_path, url=str(url), outfile=outfile))
        for infile, url in list(app.outputs.items()):
            if url.scheme in ['swift', 'swt', 'swifts', 'swts']:
                upload_cmds.append(
                    "python '{mover}' upload '{url}' '{infile}'"
                    .format(mover=mover_path, url=str(url), infile=infile))
        if download_cmds or upload_cmds:
            # copy the downloader script
            mover_src = resource_path(Requirement.parse("gc3pie"), "gc3libs/etc/mover.py")
            with open(mover_src, 'r') as src:
                with self.transport.open(mover_path, 'w') as dst:
                    dst.write(src.read())
        return download_cmds, upload_cmds

    def _read_app_process_id(self, pidfile_path):
        """
        Read and return the PID stored in `pidfile_path`.
        """
        # Just after the script has been started the pidfile should be
        # filled in with the correct pid.
        #
        # However, the script can have not been able to write the
        # pidfile yet, so we have to wait a little bit for it...
        for retry in gc3libs.utils.ExponentialBackoff():
            try:
                with self.transport.open(pidfile_path, 'r') as pidfile:
                    pid = pidfile.read().strip()
                    # it happens that the PID file exists, but `.read()`
                    # returns the empty string... just wait and retry.
                    if pid == '':
                        continue
                    return pid
            except gc3libs.exceptions.TransportError as ex:
                if '[Errno 2]' in str(ex):  # no such file or directory
                    time.sleep(retry)
                    continue
                else:
                    raise
        else:  # I wouldn't have imagined I'd ever use `for: .. else:`!
            raise gc3libs.exceptions.LRMSSubmitError(
                "Unable to read PID file of submitted process from"
                " execution directory `%s`: %s"
                % (execdir, pidfile_path))


    def update_job_state(self, app):
        """
        Query the running status of the local process whose PID is
        stored into `app.execution.lrms_jobid`, and map the POSIX
        process status to GC3Libs `Run.State`.
        """
        self._connect()
        pid = app.execution.lrms_jobid
        try:
            pstat = self._machine.get_process_state(pid)
            app.execution.state = _parse_process_status(pstat)
            if app.execution.state == Run.State.TERMINATING:
                self._cleanup_terminating_task(app, pid)
            else:
                self._kill_if_over_time_limits(app)
        except LookupError:
            log.debug(
                "Process with PID %s not found,"
                " assuming task %s has finished running.",
                pid, app)
            self._cleanup_terminating_task(app, pid)
        return app.execution.state

    def _kill_if_over_time_limits(self, app):
        pid = app.execution.lrms_jobid
        elapsed = self._machine.get_process_running_time(pid)
        # determine whether to kill, depending on wall-clock time
        cancel = False
        if elapsed > self.max_walltime:
            log.warning("Task %s has run for %s, exceeding max_walltime %s of resource %s: cancelling it.",
                        app, elapsed.to_timedelta(), self.max_walltime, self.name)
            cancel = True
        if (app.requested_walltime and elapsed > app.requested_walltime):
            log.warning("Task %s has run for %s, exceeding own `requested_walltime` %s: cancelling it.",
                        app, elapsed.to_timedelta(), app.requested_walltime)
            cancel = True
        if cancel:
            self.cancel_job(app)
            # set signal to SIGTERM in termination status
            self._cleanup_terminating_task(app, pid, termstatus=(15, -1))
            return

    def _cleanup_terminating_task(self, app, pid, termstatus=None):
        app.execution.state = Run.State.TERMINATING
        # if `self._job_infos` records this task as terminated, then
        # updates to resource utilization records has already been
        # done by `get_resource_status()`
        if (pid in self._job_infos and not self._job_infos[pid]['terminated']):
            self._job_infos[pid]['terminated'] = True
            self._write_job_info_file(pid, self._job_infos[pid])
            # do in-memory bookkeeping
            assert (self._job_infos[pid]['requested_memory'] == app.requested_memory)
            assert (self._job_infos[pid]['requested_cores'] == app.requested_cores)
            self.free_slots += app.requested_cores
            self.user_run -= 1
            if app.requested_memory is not None:
                self.available_memory += app.requested_memory
        wrapper_filename = posixpath.join(
            app.execution.lrms_execdir,
            ShellcmdLrms.PRIVATE_DIR,
            ShellcmdLrms.WRAPPER_OUTPUT_FILENAME)
        try:
            log.debug(
                "Reading resource utilization from wrapper file `%s` for task %s ...",
                wrapper_filename, app)
            with self.transport.open(wrapper_filename, 'r') as wrapper_file:
                termstatus, outcome, valid = \
                    self._parse_wrapper_output(wrapper_file, termstatus)
                if valid:
                    app.execution.update(outcome)
                if termstatus is not None:
                    app.execution.returncode = termstatus
                else:
                    app.execution.returncode = (Run.Signals.RemoteError, -1)
        except EnvironmentError as err:
            msg = ("Could not read wrapper file `{0}` for task `{1}`: {2}"
                   .format(wrapper_filename, app, err))
            log.warning("%s -- Termination status and resource utilization fields will not be set.", msg)
            raise gc3libs.exceptions.InvalidValue(msg)

    def _parse_wrapper_output(self, wrapper_file, termstatus=None):
        """
        Parse the file saved by the wrapper in
        `ShellcmdLrms.WRAPPER_OUTPUT_FILENAME` inside the PRIVATE_DIR
        in the job's execution directory and return a `Struct`:class:
        containing the values found on the file.

        `wrapper_file` is an opened file. This method will rewind the
        file before reading.
        """
        valid = True  # optimistic default
        wrapper_file.seek(0)
        acctinfo = Struct()
        for line in wrapper_file:
            line = line.strip()
            # if the executed command was killed by a signal, then GNU
            # `time` will output `0` for the `%x` format specifier but
            # still mark the exit signal in the output.
            if line.startswith('Command terminated by signal'):
                # since `Run.shellexit_to_returncode()` returns a
                # pair, we need to use a pair as default for
                # `returncode` here as well, to avoid comparing an
                # integer with a tuple in the `max()` invocation
                # at line 1670 below.
                termstatus = (int(line.split()[4]), -1)
                continue
            if '=' not in line:
                continue
            k, v = line.split('=', 1)
            if k not in self.TIMEFMT_CONV:
                gc3libs.log.warning(
                    "Unknown key '%s' in wrapper output file - ignoring!", k)
                continue
            name, conv = self.TIMEFMT_CONV[k]
            # the `time` man page states that: "Any character
            # following a percent sign that is not listed in the table
            # below causes a question mark (`?') to be output [...] to
            # indicate that an invalid resource specifier was given."
            # Actually, `time` seems to print a question mark also
            # when a value is not available (e.g., corresponding data
            # not available/collected by the kernel) so we just set a
            # field to ``None`` if there is a question mark in it.
            if v.startswith('?'):
                acctinfo[name] = None
            elif name == 'returncode':
                # be sure not to overwrite here the error exit set in
                # lines 1643--1645 above with the successful exit
                # reported by GNU time's `%x` specifier
                if termstatus is not None:
                    termstatus = max(termstatus, conv(v))
                else:
                    termstatus = conv(v)
            else:
                acctinfo[name] = conv(v)

        # apprently GNU time does not report the total CPU time, used
        # so compute it here
        try:
            acctinfo['used_cpu_time'] = (
                acctinfo['shellcmd_user_time'] + acctinfo['shellcmd_kernel_time'])
        except KeyError:
            valid = False

        return termstatus, acctinfo, valid


    def validate_data(self, data_file_list=[]):
        """
        Return `False` if any of the URLs in `data_file_list` cannot
        be handled by this backend.

        The `shellcmd`:mod: backend can handle the following URL schemas:

        - ``file`` (natively, read/write);
        - ``swift``/``swifts``/``swt``/``swts`` (with Python-based remote helper, read/write);
        - ``http``/``https`` (with Python-based remote helper, read-only).
        """
        for url in data_file_list:
            if url.scheme not in ['file', 'swift', 'swifts', 'swt', 'swts', 'http', 'https']:
                return False
        return True


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#

"""
Test that classes from `gc3libs.quantity.Quantity` behave well
also when `from __future__ import division` is in effect.

Se issue #525 for background and possibly more info.
"""
# Copyright (C) 2012, 2013,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from __future__ import division
__docformat__ = 'reStructuredText'


from random import randint

# GC3Pie imports
from gc3libs.quantity import Duration, Memory


# test definitions

def test_divide_duration1():
    n = randint(1, 100)
    d1 = Duration(2*n, unit=Duration.s)
    d2 = d1 / 2
    assert d2 ==   Duration(n, unit=Duration.s)
    assert 2*d2 == d1
    assert d2*2 == d1

# same, but with a non-base unit
def test_divide_duration2():
    n = randint(1, 100)
    d1 = Duration(2*n, unit=Duration.days)
    d2 = d1 / 2
    assert d2 ==   Duration(n, unit=Duration.days)
    assert 2*d2 == d1
    assert d2*2 == d1


def test_divide_memory1():
    n = randint(1, 100)
    d1 = Memory(2*n, unit=Memory.B)
    d2 = d1 / 2
    assert d2 ==   Memory(n, unit=Memory.B)
    assert 2*d2 == d1
    assert d2*2 == d1


# same, but with a non-base unit
def test_divide_memory2():
    n = randint(1, 100)
    d1 = Memory(2*n, unit=Memory.MB)
    d2 = d1 / 2
    assert d2 ==   Memory(n, unit=Memory.MB)
    assert 2*d2 == d1
    assert d2*2 == d1


# main: run tests

if "__main__" == __name__:
    import pytest
    pytest.main(["-v", __file__])



# test_core.py
# -*- coding: utf-8 -*-
#
#  Copyright (C) 2015-2018  University of Zurich. All rights reserved.
#
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# stdlib imports
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
import os

import pytest

# GC3Pie imports
from gc3libs import Run, Application, create_core
import gc3libs.config
from gc3libs.core import Core, MatchMaker
from gc3libs.quantity import GB, hours
from gc3libs.utils import string_to_boolean

from gc3libs.testing.helpers import temporary_config_file, temporary_core


def test_core_resources():
    """
    Check that configured resources can be accessed through the `Core` object.
    """
    with temporary_core() as core:
        resources = core.resources
        assert len(resources) == 1
        assert 'test' in resources
        test_rsc = resources['test']
        # these should match the resource definition in `gc3libs.testing.helpers.temporary_core`
        assert test_rsc.max_cores_per_job == 123
        assert test_rsc.max_memory_per_core == 999*GB
        assert test_rsc.max_walltime == 7*hours


@pytest.mark.skipif(
    string_to_boolean(os.environ.get('GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL', 'no')),
    reason="Skipping test: not compatible with GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL=yes")
def test_core_disable_resource_on_auth_init_failure():
    """Test that a resource is disabled if the auth cannot be initialized successfully."""

    # pylint: disable=no-self-use,unused-argument
    class BadInitAuth(object):
        """Fail all authentication methods."""
        def __init__(self, **auth):
            raise RuntimeError("Bad authentication object!")

        def is_valid(self):
            raise AssertionError("This method should have never been called!")

        def check(self):
            raise AssertionError("This method should have never been called!")

        def enable(self):
            raise AssertionError("This method should have never been called!")

    with pytest.raises(gc3libs.exceptions.NoResources):
        _test_core_disable_resource_on_auth_failure(BadInitAuth)


@pytest.mark.skipif(
    string_to_boolean(os.environ.get('GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL', 'no')),
    reason="Skipping test: not compatible with GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL=yes")
def test_core_disable_resource_on_auth_check_failure():
    """Test that a resource is disabled if the auth cannot be checked successfully."""

    # pylint: disable=no-self-use,unused-argument
    class BadCheckAuth(object):
        """Fail `Authentication.check()`"""

        def __init__(self, **auth):
            pass

        def is_valid(self):
            raise AssertionError("This method should have never been called!")

        def check(self):
            raise RuntimeError("Bad authentication object!")

        def enable(self):
            raise AssertionError("This method should have never been called!")

    with pytest.raises(gc3libs.exceptions.NoResources):
        _test_core_disable_resource_on_auth_failure(BadCheckAuth)


@pytest.mark.skipif(
    string_to_boolean(os.environ.get('GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL', 'no')),
    reason="Skipping test: not compatible with GC3PIE_RESOURCE_INIT_ERRORS_ARE_FATAL=yes")
def test_core_disable_resource_on_auth_enable_failure():
    """Test that a resource is disabled if the auth cannot be enabled successfully."""

    # pylint: disable=no-self-use,unused-argument
    class BadEnableAuth(object):
        """Fail `Authentication.enable()`"""

        def __init__(self, **auth):
            pass

        def is_valid(self):
            raise AssertionError("This method should have never been called!")

        def check(self):
            return False  # so that `enable` is called next

        def enable(self):
            raise RuntimeError("Bad authentication object!")

    with pytest.raises(gc3libs.exceptions.NoResources):
        _test_core_disable_resource_on_auth_failure(BadEnableAuth)


def _test_core_disable_resource_on_auth_failure(auth_cls):
    """Common code for `test_core_disable_resource_on_auth_*_failure`."""
    gc3libs.authentication.Auth.register('bad', auth_cls)
    # set up
    cfg = gc3libs.config.Configuration()
    cfg.auths['bad_auth'].update(
        type='bad',
        username='fake',
    )
    cfg.resources['test'].update(
        name='test',
        type='shellcmd',
        transport='ssh',
        auth='bad_auth',
        max_cores_per_job=1,
        max_memory_per_core=1*GB,
        max_walltime=8*hours,
        max_cores=10,
        architecture=Run.Arch.X86_64,
    )
    core = Core(cfg)


def test_create_core_default():
    """Test `create_core` with factory defaults."""
    with temporary_config_file() as cfgfile:
        # std factory params
        core = create_core(cfgfile.name)
        assert isinstance(core, Core)
        assert core.auto_enable_auth == True


def test_create_core_non_default():
    """Test `create_core` with non-default arguments."""
    with temporary_config_file() as cfgfile:
        # use a specific MatchMaker instance for equality testing
        mm = MatchMaker()
        core = create_core(cfgfile.name, matchmaker=mm)
        assert core.auto_enable_auth == True
        assert core.matchmaker == mm


def test_create_core_no_auto_enable_auth():
    """Test `create_core` without the "auto enable" feature."""
    with temporary_config_file() as cfgfile:
        # std factory params
        core = create_core(cfgfile.name, auto_enable_auth=False)
        assert core.auto_enable_auth == False
